# Authors: Eric Larson <larson.eric.d@gmail.com>
#
# License: BSD (3-clause)

import contextlib
import inspect

import numpy as np
from mne import Info, pick_info, pick_types
from mne.channels import make_standard_montage
from mne.channels.montage import transform_to_head
from mne.transforms import _get_trans, apply_trans
from mne.utils import _check_option, _validate_type, logger, verbose
from mne.viz import Brain


@verbose
def plot_3d_montage(
    info,
    view_map,
    *,
    src_det_names="auto",
    ch_names="numbered",
    subject="fsaverage",
    trans="fsaverage",
    surface="pial",
    subjects_dir=None,
    verbose=None,
):
    """
    Plot a 3D sensor montage.

    Parameters
    ----------
    info : instance of Info
        Measurement info.
    view_map : dict
        Dict of view (key) to channel-pair-numbers (value) to use when
        plotting. Note that, because these get plotted as 1-based channel
        *numbers*, the values should be 1-based rather than 0-based.
        The keys are of the form:

        ``'{side}-{view}'``
            For views like ``'left-lat'`` or ``'right-frontal'`` where the side
            matters.
        ``'{view}'``
            For views like ``'caudal'`` that are along the midline.

        See :meth:`mne.viz.Brain.show_view` for ``view`` options, and the
        Examples section below for usage examples.
    src_det_names : None | dict | str
        Source and detector names to use. "auto" (default) will see if the
        channel locations correspond to standard 10-20 locations and will
        use those if they do (otherwise will act like None). None will use
        S1, S2, ..., D1, D2, ..., etc. Can also be an explicit dict mapping,
        for example::

            src_det_names=dict(S1='Fz', D1='FCz', ...)
    ch_names : str | dict | None
        If ``'numbered'`` (default), use ``['1', '2', ...]`` for the channel
        names, or ``None`` to use ``['S1_D2', 'S2_D1', ...]``. Can also be a
        dict to provide a mapping from the ``'S1_D2'``-style names (keys) to
        other names, e.g., ``defaultdict(lambda: '')`` will prevent showing
        the names altogether.

        .. versionadded:: 0.3
    subject : str
        The subject.
    trans : str | Transform
        The subjects head<->MRI transform.
    surface : str
        The FreeSurfer surface name (e.g., 'pial', 'white').
    subjects_dir : str
        The subjects directory.
    %(verbose)s

    Returns
    -------
    figure : matplotlib.figure.Figure
        The matplotlib figimage.

    Examples
    --------
    For a Hitachi system with two sets of 12 source-detector arrangements,
    one on each side of the head, showing 1-12 on the left and 13-24 on the
    right can be accomplished using the following ``view_map``::

        >>> view_map = {
        ...     'left-lat': np.arange(1, 13),
        ...     'right-lat': np.arange(13, 25),
        ... }

    NIRx typically involves more complicated arrangements. See
    :ref:`the 3D tutorial <tut-fnirs-vis-brain-plot-3d-montage>` for
    an advanced example that incorporates the ``'caudal'`` view as well.
    """  # noqa: E501
    import matplotlib.pyplot as plt
    from scipy.spatial.distance import cdist

    _validate_type(info, Info, "info")
    _validate_type(view_map, dict, "views")
    _validate_type(src_det_names, (None, dict, str), "src_det_names")
    _validate_type(ch_names, (dict, str, None), "ch_names")
    info = pick_info(info, pick_types(info, fnirs=True, exclude=())[::2])
    if isinstance(ch_names, str):
        _check_option("ch_names", ch_names, ("numbered",), extra="when str")
        ch_names = {
            name.split()[0]: str(ni) for ni, name in enumerate(info["ch_names"], 1)
        }
    info["bads"] = []
    if isinstance(src_det_names, str):
        _check_option("src_det_names", src_det_names, ("auto",), extra="when str")
        # Decide if we can map to 10-20 locations
        names, pos = zip(
            *transform_to_head(make_standard_montage("standard_1020"))
            .get_positions()["ch_pos"]
            .items()
        )
        pos = np.array(pos, float)
        locs = dict()
        bad = False
        for ch in info["chs"]:
            name = ch["ch_name"]
            s_name, d_name = name.split()[0].split("_")
            for name, loc in [(s_name, ch["loc"][3:6]), (d_name, ch["loc"][6:9])]:
                if name in locs:
                    continue
                # see if it's close enough
                idx = np.where(cdist(loc[np.newaxis], pos)[0] < 1e-3)[0]
                if len(idx) < 1:
                    bad = True
                    break
                # Some are duplicated (e.g., T7+T3) but we can rely on the
                # first one being the canonical one
                locs[name] = names[idx[0]]
            if bad:
                break
        if bad:
            src_det_names = None
            logger.info(
                "Could not automatically map source/detector names to 10-20 locations."
            )
        else:
            src_det_names = locs
            logger.info("Source-detector names automatically mapped to 10-20 locations")

    head_mri_t = _get_trans(trans, "head", "mri")[0]
    del trans
    views = list()
    for key, num in view_map.items():
        _validate_type(key, str, f"view_map key {repr(key)}")
        _validate_type(num, np.ndarray, f"view_map[{repr(key)}]")
        if "-" in key:
            hemi, v = key.split("-", maxsplit=1)
            hemi = dict(left="lh", right="rh")[hemi]
            views.append((hemi, v, num))
        else:
            views.append(("lh", key, num))
    del view_map
    size = (400 * len(views), 400)
    brain = Brain(
        subject,
        "both",
        surface,
        views=["lat"] * len(views),
        size=size,
        background="w",
        units="m",
        view_layout="horizontal",
        subjects_dir=subjects_dir,
    )
    with _safe_brain_close(brain):
        brain.add_head(dense=False, alpha=0.1)
        brain.add_sensors(
            info, trans=head_mri_t, fnirs=["channels", "pairs", "sources", "detectors"]
        )
        add_text_kwargs = dict()
        if "render" in inspect.signature(brain.plotter.add_text).parameters:
            add_text_kwargs["render"] = False
        for col, view in enumerate(views):
            plotted = set()
            brain.show_view(
                view[1],
                hemi=view[0],
                focalpoint=(0, -0.02, 0.02),
                distance=0.4,
                row=0,
                col=col,
            )
            brain.plotter.subplot(0, col)
            vp = brain.plotter.renderer
            for ci in view[2]:  # figure out what we need to plot
                this_ch = info["chs"][ci - 1]
                ch_name = this_ch["ch_name"].split()[0]
                s_name, d_name = ch_name.split("_")
                needed = [
                    (ch_names, "ch_names", ch_name, this_ch["loc"][:3], 12, "Centered"),
                    (
                        src_det_names,
                        "src_det_names",
                        s_name,
                        this_ch["loc"][3:6],
                        8,
                        "Bottom",
                    ),
                    (
                        src_det_names,
                        "src_det_names",
                        d_name,
                        this_ch["loc"][6:9],
                        8,
                        "Bottom",
                    ),
                ]
                for lookup, lname, name, ch_pos, font_size, va in needed:
                    if name in plotted:
                        continue
                    plotted.add(name)
                    orig_name = name
                    if lookup is not None:
                        name = lookup[name]
                    _validate_type(name, str, f"{lname}[{repr(orig_name)}]")
                    ch_pos = apply_trans(head_mri_t, ch_pos)
                    vp.SetWorldPoint(np.r_[ch_pos, 1.0])
                    vp.WorldToDisplay()
                    ch_pos = np.array(vp.GetDisplayPoint()[:2]) - np.array(
                        vp.GetOrigin()
                    )
                    actor = brain.plotter.add_text(
                        name,
                        ch_pos,
                        font_size=font_size,
                        color=(0.0, 0.0, 0.0),
                        **add_text_kwargs,
                    )
                    prop = actor.GetTextProperty()
                    getattr(prop, f"SetVerticalJustificationTo{va}")()
                    prop.SetJustificationToCentered()
                    actor.SetTextProperty(prop)
                    prop.SetBold(True)
        img = brain.screenshot()
    return plt.figimage(img, resize=True).figure


@contextlib.contextmanager
def _safe_brain_close(brain):
    try:
        yield
    finally:
        brain.close()



# Authors: Robert Luke <mail@robertluke.net>
#
# License: BSD (3-clause)

import numpy as np
import pandas as pd
from mne import pick_types
from mne.io import BaseRaw
from mne.utils import _require_version, _validate_type


def quantify_mayer_fooof(
    raw,
    num_oscillations=1,
    centre_frequency=0.01,
    extra_df_fields=None,
    fmin=0.001,
    fmax=1,
    tmin=0,
    tmax=None,
    n_fft=400,
    n_overlap=200,
    peak_width_limits=(0.5, 12.0),
):
    """
    Quantify Mayer wave properties using FOOOF analysis.

    The Fitting Oscillations & One Over F (FOOOF)
    :footcite:`donoghue2020parameterizing`
    is utilised to estimate Mayer wave oscillation parameters as described in
    :footcite:`luke2021characterization`.

    The FOOOF algorithm is applied to the mean PSD estimate of the data,
    the oscillation closest to the `centre_frequency` is assumed to be the
    Mayer wave oscillation. The parameters for this oscillation are returned
    as a dataframe. You can return multiple closest oscillations to the
    centre_frequency by increasing the `num_oscillations` parameter.

    Parameters
    ----------
    raw : instance of Raw
        The haemoglobin data.
    num_oscillations : number
        Number of parameterised oscilations to be returned. These are selected
        in increasing distance from the `centre_frequency`.
    centre_frequency : number
        Centre frequency of the Mayer wave.
    extra_df_fields : number
        Dictionary of values to be appended to the dataframe.
    fmin : float
        Min frequency of interest.
    fmax : float
        Max frequency of interest.
    tmin : float | None
        Min time of interest.
    tmax : float | None
        Max time of interest.
    n_fft : int
        The length of FFT used, must be ``>= n_per_seg`` (default: 256).
        The segments will be zero-padded if ``n_fft > n_per_seg``.
        If n_per_seg is None, n_fft must be <= number of time points
        in the data.
    n_overlap : int
        The number of points of overlap between segments. Will be adjusted
        to be <= n_per_seg. The default value is 0.
    peak_width_limits : tuple of (float, float), optional, default: (0.5, 12.0)
        Limits on possible peak width, in Hz, as (lower_bound, upper_bound).
        As used by FOOOF.

    Returns
    -------
    df : DataFrame
        Dataframe with columns.

    References
    ----------
    .. footbibliography::
    """
    _require_version("fooof", "run the FOOOF algorithm.")
    _validate_type(raw, BaseRaw, "raw")

    extra_df_fields = {} if extra_df_fields is None else extra_df_fields
    hbo_picks = pick_types(raw.info, fnirs="hbo")
    hbr_picks = pick_types(raw.info, fnirs="hbr")

    if (not len(hbo_picks)) & (not len(hbr_picks)):
        # It may be perfectly valid to compute this on optical density
        # or raw data, I just haven't tried this. Let me know if this works
        # for you and we can ease this restriction.
        raise RuntimeError(
            "Mayer wave estimation should be run on haemoglobin concentration data."
        )

    df = pd.DataFrame()

    for picks, chroma in zip([hbo_picks, hbr_picks], ["hbo", "hbr"]):
        if len(picks):
            fm_hbo = _run_fooof(
                raw.copy().pick(picks),
                fmin=fmin,
                fmax=fmax,
                tmin=tmin,
                tmax=tmax,
                n_overlap=n_overlap,
                n_fft=n_fft,
                peak_width_limits=peak_width_limits,
            )

            cf, pw, bw = _process_fooof_output(fm_hbo, centre_frequency)

            data = dict()
            data["Centre Frequency"] = cf
            data["Bandwidth"] = bw
            data["Power"] = pw
            data["Chromaphore"] = chroma
            data = {**data, **extra_df_fields}

            df = pd.concat([df, pd.DataFrame(data, index=[0])], ignore_index=True)

    return df


def _run_fooof(
    raw,
    fmin=0.001,
    fmax=1,
    tmin=0,
    tmax=None,
    n_overlap=200,
    n_fft=400,
    peak_width_limits=(0.5, 12.0),
):
    """Prepare data for FOOOF including welch and scaling, then apply."""
    from fooof import FOOOF

    psd = raw.compute_psd(
        fmin=fmin, fmax=fmax, tmin=tmin, tmax=tmax, n_overlap=n_overlap, n_fft=n_fft
    )
    spectra, freqs = psd.get_data(return_freqs=True)

    # FOOOF doesn't like low frequencies, so multiple by 10.
    # This is corrected for in the output below.
    freqs = freqs * 10

    # Remember these values are really 0.001 to 1.2 Hz
    fm = FOOOF(peak_width_limits=peak_width_limits)

    # And these values are really 0.0001 to 1 Hz
    freq_range = [0.001, 10]

    fm.fit(freqs, np.mean(spectra, axis=0), freq_range)

    return fm


def _process_fooof_output(fm, centre_frequency):
    """Extract and post fix FOOOF result."""
    CFs = [d[0] for d in fm.peak_params_]

    # In this line we correct for the scaling done in _run_fooof()
    mayer_idx = _find_nearest_idx(CFs, centre_frequency * 10)

    mayer = fm.peak_params_[mayer_idx]

    cf = mayer[0] / 10  # Correct scaling
    bw = mayer[2] / 10  # Correct scaling
    pw = mayer[1]

    return cf, pw, bw


def _find_nearest_idx(a, a0):
    """Element idx in nd array `a` closest to the scalar value `a0`."""
    if isinstance(a, list):
        a = np.array(a)
    idx = np.abs(a - a0).argmin()
    return idx



# Authors: Robert Luke <mail@robertluke.net>
#
# License: BSD (3-clause)

import mne
from mne.io import BaseRaw
from mne.preprocessing.nirs import source_detector_distances
from mne.utils import _validate_type


def get_short_channels(raw, max_dist=0.01):
    """
    Return channels with a short source-detector separation.

    Parameters
    ----------
    raw : instance of Raw
        Raw instance containing fNIRS data.
    max_dist : number
        Maximum distance of returned channels (m).

    Returns
    -------
    raw : instance of Raw
        Raw instance with only short channels.
    """
    short_chans = raw.copy().load_data()
    _validate_type(short_chans, BaseRaw, "raw")

    picks = mne.pick_types(
        short_chans.info, meg=False, eeg=False, fnirs=True, exclude=[]
    )
    if not len(picks):
        raise RuntimeError("Short channel extraction for NIRS signals only.")

    dists = source_detector_distances(short_chans.info, picks=picks)
    short_chans.pick(picks[dists < max_dist])

    return short_chans


def get_long_channels(raw, min_dist=0.015, max_dist=0.045):
    """
    Return channels with a long source detector separation.

    Parameters
    ----------
    raw : instance of Raw
        Raw instance containing fNIRS data.
    min_dist : number
        Minimum distance of returned channels (m).
    max_dist : number
        Maximum distance of returned channels (m).

    Returns
    -------
    raw : instance of Raw
        Raw instance with only long channels.
    """
    long_chans = raw.copy().load_data()
    _validate_type(long_chans, BaseRaw, "raw")

    picks = mne.pick_types(
        long_chans.info, meg=False, eeg=False, fnirs=True, exclude=[]
    )
    if not len(picks):
        raise RuntimeError("Short channel extraction for NIRS signals only.")

    dists = source_detector_distances(long_chans.info, picks=picks)
    long_chans.pick(picks[(dists > min_dist) & (dists < max_dist)])

    return long_chans



# Authors: Robert Luke <mail@robertluke.net>
#
# License: BSD (3-clause)

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import seaborn as sns


def plot_timechannel_quality_metric(raw, scores, times, threshold=0.1, title=None):
    """
    Plot time x channel based quality metrics.

    The left figure shows the raw score per channel and time.

    Parameters
    ----------
    raw : instance of Raw
        The haemoglobin data.
    scores : array
        The quality metric scores.
    times : list of pairs
        Start and end time for each quality metric.
    threshold : float
        Value below which a segment will be marked as bad.
    title : str
        Title of plot. If not specified a default title will be used.

    Returns
    -------
    fig : figure
        Matplotlib figure displaying raw scores and thresholded scores.
    """
    ch_names = raw.ch_names
    cols = [np.round(t[0]) for t in times]

    if title is None:
        title = "Automated noisy channel detection: fNIRS"

    data_to_plot = pd.DataFrame(
        data=scores,
        columns=pd.Index(cols, name="Time (s)"),
        index=pd.Index(ch_names, name="Channel"),
    )

    n_chans = len(ch_names)
    vsize = 0.2 * n_chans

    # First, plot the "raw" scores.
    fig, ax = plt.subplots(1, 2, figsize=(20, vsize), layout="constrained")
    fig.suptitle(title, fontsize=16, fontweight="bold")
    sns.heatmap(
        data=data_to_plot,
        cmap="Reds_r",
        vmin=0,
        vmax=1,
        cbar_kws=dict(label="Score"),
        ax=ax[0],
    )
    [
        ax[0].axvline(x, ls="dashed", lw=0.25, dashes=(25, 15), color="gray")
        for x in range(1, len(times))
    ]
    ax[0].set_title("All Scores", fontweight="bold")
    markbad(raw, ax[0])

    # Now, adjust the color range to highlight segments that exceeded the
    # limit.

    data_to_plot = pd.DataFrame(
        data=scores > threshold,
        columns=pd.Index(cols, name="Time (s)"),
        index=pd.Index(ch_names, name="Channel"),
    )
    sns.heatmap(
        data=data_to_plot,
        vmin=0,
        vmax=1,
        cmap="Reds_r",
        cbar_kws=dict(label="Score"),
        ax=ax[1],
    )
    [
        ax[1].axvline(x, ls="dashed", lw=0.25, dashes=(25, 15), color="gray")
        for x in range(1, len(times))
    ]
    ax[1].set_title("Scores < Limit", fontweight="bold")
    markbad(raw, ax[1])

    return fig


def markbad(raw, ax):
    [
        ax.axhline(y + 0.5, ls="solid", lw=2, color="black")
        for y in np.where([ch in raw.info["bads"] for ch in raw.ch_names])[0]
    ]

    return ax



# Authors: Robert Luke <mail@robertluke.net>
#
# License: BSD (3-clause)

from ._peak_power import peak_power
from ._scalp_coupling_segmented import scalp_coupling_index_windowed
from ._mayer import quantify_mayer_fooof



# Author: Robert Luke <mail@robertluke.net>
#
# License: BSD (3-clause)

import os.path as op

import mne
import numpy as np
import pandas as pd
from mne.io import BaseRaw
from mne.io.constants import FIFF
from mne.transforms import _get_trans, apply_trans
from mne.utils import _check_fname, _validate_type, warn


def _read_fold_xls(fname, atlas="Juelich"):
    """Read fOLD toolbox xls file.

    The values are then manipulated in to a tidy dataframe.

    Note the xls files are not included as no license is provided.

    Parameters
    ----------
    fname : str
        Path to xls file.
    atlas : str
        Requested atlas.
    """
    page_reference = {"AAL2": 2, "AICHA": 5, "Brodmann": 8, "Juelich": 11, "Loni": 14}

    tbl = pd.read_excel(fname, sheet_name=page_reference[atlas])

    # Remove the spacing between rows
    empty_rows = np.where(np.isnan(tbl["Specificity"]))[0]
    tbl = tbl.drop(empty_rows).reset_index(drop=True)

    # Empty values in the table mean its the same as above
    for row_idx in range(1, tbl.shape[0]):
        for col_idx, col in enumerate(tbl.columns):
            if not isinstance(tbl[col][row_idx], str):
                if np.isnan(tbl[col][row_idx]):
                    tbl.iloc[row_idx, col_idx] = tbl.iloc[row_idx - 1, col_idx]

    tbl["Specificity"] = tbl["Specificity"] * 100
    tbl["brainSens"] = tbl["brainSens"] * 100
    return tbl


def _generate_montage_locations():
    """Get standard MNI montage locations in dataframe.

    Data is returned in the same format as the eeg_positions library.
    """
    # standard_1020 and standard_1005 are in MNI (fsaverage) space already,
    # but we need to undo the scaling that head_scale will do
    montage = mne.channels.make_standard_montage(
        "standard_1005", head_size=0.09700884729534559
    )
    for d in montage.dig:
        d["coord_frame"] = FIFF.FIFFV_MNE_COORD_MNI_TAL
    montage.dig[:] = montage.dig[3:]
    montage.add_mni_fiducials()  # now in fsaverage space
    coords = pd.DataFrame.from_dict(montage.get_positions()["ch_pos"]).T
    coords["label"] = coords.index
    coords = coords.rename(columns={0: "x", 1: "y", 2: "z"})

    return coords.reset_index(drop=True)


def _find_closest_standard_location(position, reference, *, out="label"):
    """Return closest montage label to coordinates.

    Parameters
    ----------
    position : array, shape (3,)
        Coordinates.
    reference : dataframe
        As generated by _generate_montage_locations.
    trans_pos : str
        Apply a transformation to positions to specified frame.
        Use None for no transformation.
    """
    from scipy.spatial.distance import cdist

    p0 = np.array(position)
    p0.shape = (-1, 3)
    head_mri_t, _ = _get_trans("fsaverage", "head", "mri")
    p0 = apply_trans(head_mri_t, p0)
    dists = cdist(p0, np.asarray(reference[["x", "y", "z"]], float))

    if out == "label":
        min_idx = np.argmin(dists)
        return reference["label"][min_idx]
    else:
        assert out == "dists"
        return dists


def fold_landmark_specificity(
    raw, landmark, fold_files=None, atlas="Juelich", interpolate=False
):
    """Return the specificity of each channel to a specified brain landmark.

    Parameters
    ----------
    raw : BaseRaw
        The fNIRS data.
    landmark : str
        Landmark of interest. Must be present in fOLD toolbox data file.
    fold_files : list | path-like | None
        If None, will use the MNE_NIRS_FOLD_PATH config variable.
        If path-like, should be a path to a directory containing '10-10.xls'
        and '10-5.xls'. If list, should be paths to the fold toolbox files.
        See the Notes section of :func:`~mne_nirs.io.fold_channel_specificity`
        for details.
    atlas : str
        Brain atlas to use.
    interpolate : bool
        If the optimal source-detector pair is not found in the fOLD files
        False (default) will yield no results for that pairing, whereas
        True will use the next closest match. See Notes of
        :func:`mne_nirs.io.fold_channel_specificity` for an example.

        .. warning::
           The sensitivity profile can differ substantially for nearest
           neighbors, so use ``interpolate=True`` with caution.

    Returns
    -------
    spec : array
        Specificity values for each channel to brain landmark.

    See Also
    --------
    fold_landmark_specificity

    Notes
    -----
    Specificity values are provided by the fOLD toolbox
    :footcite:`morais2018fnirs` excel files. See the Notes section of
    :func:`~mne_nirs.io.fold_channel_specificity` for more details.

    References
    ----------
    .. footbibliography::
    """
    _validate_type(landmark, str, "landmark")
    _validate_type(raw, BaseRaw, "raw")

    reference_locations = _generate_montage_locations()

    fold_tbl = _check_load_fold(fold_files, atlas)

    specificity = np.zeros(len(raw.ch_names))
    for cidx in range(len(raw.ch_names)):
        tbl = _source_detector_fold_table(
            raw, cidx, reference_locations, fold_tbl, interpolate
        )

        if len(tbl) > 0:
            tbl["ContainsLmk"] = [landmark in la for la in tbl["Landmark"]]
            tbl = tbl.query("ContainsLmk == True")["Specificity"]

        if len(tbl) == 0:
            continue
            # print(f"No data for {src_name}-{det_name}")
        elif len(tbl) == 1:
            specificity[cidx] = tbl.values[0]
        else:
            raise RuntimeError("Multiple specificity values returned")

    return np.array(specificity)


def fold_channel_specificity(raw, fold_files=None, atlas="Juelich", interpolate=False):
    """Return the landmarks and specificity a channel is sensitive to.

    Parameters
    ----------
    raw : BaseRaw
        The fNIRS data.
    fold_files : list | path-like | None
        If None, will use the MNE_NIRS_FOLD_PATH config variable.
        If path-like, should be a path to a directory containing '10-10.xls'
        and '10-5.xls'. If list, should be paths to the fold toolbox files.
        See Notes for details.
    atlas : str
        Brain atlas to use.
    interpolate : bool
        If the optimal source-detector pair is not found in the fOLD files
        False (default) will yield no results for that pairing, whereas
        True will use the next closest match. See Notes for an example.

        .. warning:: The sensitivity profile can differ substantially for
                     nearest neighbors, so use ``interpolate=True`` with
                     caution.

    Returns
    -------
    spec : list of DataFrame
        List of dataframes, one for each channel.

    See Also
    --------
    fold_landmark_specificity

    Notes
    -----
    **fOLD Toolbox**

    Specificity values are provided by the fOLD toolbox
    :footcite:`morais2018fnirs` excel files.
    For licensing reasons, these files are not distributed with MNE-NIRS.
    You need to download them from
    `the author's website <https://github.com/nirx/fOLD-public>`__.
    To automatically utilize the ``MNE_NIRS_FOLD_PATH`` config for the
    ``fold_files`` parameter, you can download the entire ``fOLD-public``
    repository `as a zip <https://github.com/nirx/fOLD-public/archive/refs/heads/master.zip>`__
    and expand it to some suitable location like
    ``~/mne_data/fOLD/fOLD-public-master``, and then set the config value
    on your machine by using :func:`mne:mne.set_config` like::

        >>> mne.set_config('MNE_NIRS_FOLD_PATH', '~/mne_data/fOLD/fOLD-public-master/Supplementary')

    From then on, :func:`~mne_nirs.io.fold_channel_specificity` and
    :func:`~mne_nirs.io.fold_landmark_specificity` will automatically use this
    directory to find the fOLD xls files when you pass ``fold_files=None``
    (which is the default). We recommend following this procedure so that
    the files can be reused automatically.

    **Interpolation**

    For an example of interpolation, consider the pairings P5-PO7 nor P6-PO8,
    neither of which are listed in the fOLD toolbox:

    - With ``interpolate=False``, the returned ``spec`` will not have entries
      for these channels. When
    - With ``interpolate=True``, entries like P7-PO7 and P8-PO8, respectively,
      might be used instead. A warning is emitted if such substitutions are
      made.

    References
    ----------
    .. footbibliography::
    """  # noqa: E501
    _validate_type(raw, BaseRaw, "raw")

    reference_locations = _generate_montage_locations()

    fold_tbl = _check_load_fold(fold_files, atlas)

    chan_spec = list()
    for cidx in range(len(raw.ch_names)):
        tbl = _source_detector_fold_table(
            raw, cidx, reference_locations, fold_tbl, interpolate
        )
        chan_spec.append(tbl.reset_index(drop=True))

    return chan_spec


def _check_load_fold(fold_files, atlas):
    _validate_type(fold_files, (list, "path-like", None), "fold_files")
    if fold_files is None:
        fold_files = mne.get_config("MNE_NIRS_FOLD_PATH")
        if fold_files is None:
            raise ValueError(
                "MNE_NIRS_FOLD_PATH not set, either set it using "
                "mne.set_config or pass fold_files as str or list"
            )
    if not isinstance(fold_files, list):  # path-like
        fold_files = _check_fname(
            fold_files,
            overwrite="read",
            must_exist=True,
            name="fold_files",
            need_dir=True,
        )
        fold_files = [op.join(fold_files, f"10-{x}.xls") for x in (5, 10)]

    fold_tbl = pd.DataFrame()
    for fi, fname in enumerate(fold_files):
        fname = _check_fname(
            fname, overwrite="read", must_exist=True, name=f"fold_files[{fi}]"
        )
        fold_tbl = pd.concat(
            [fold_tbl, _read_fold_xls(fname, atlas=atlas)], ignore_index=True
        )
    return fold_tbl


def _source_detector_fold_table(raw, cidx, reference, fold_tbl, interpolate):
    src = raw.info["chs"][cidx]["loc"][3:6]
    det = raw.info["chs"][cidx]["loc"][6:9]

    ref_lab = list(reference["label"])
    dists = _find_closest_standard_location([src, det], reference, out="dists")
    src_min, det_min = np.argmin(dists, axis=1)
    src_name, det_name = ref_lab[src_min], ref_lab[det_min]

    tbl = fold_tbl.query("Source == @src_name and Detector == @det_name")
    dist = np.linalg.norm(dists[[0, 1], [src_min, det_min]])
    # Try reversing source and detector
    if len(tbl) == 0:
        tbl = fold_tbl.query("Source == @det_name and Detector == @src_name")
    if len(tbl) == 0 and interpolate:
        # Try something hopefully not too terrible: pick the one with the
        # smallest net distance
        good = np.isin(fold_tbl["Source"], reference["label"]) & np.isin(
            fold_tbl["Detector"], reference["label"]
        )
        assert good.any()
        tbl = fold_tbl[good]
        assert len(tbl)
        src_idx = [ref_lab.index(src) for src in tbl["Source"]]
        det_idx = [ref_lab.index(det) for det in tbl["Detector"]]
        # Original
        tot_dist = np.linalg.norm([dists[0, src_idx], dists[1, det_idx]], axis=0)
        assert tot_dist.shape == (len(tbl),)
        idx = np.argmin(tot_dist)
        dist_1 = tot_dist[idx]
        src_1, det_1 = ref_lab[src_idx[idx]], ref_lab[det_idx[idx]]
        # And the reverse
        tot_dist = np.linalg.norm([dists[0, det_idx], dists[1, src_idx]], axis=0)
        idx = np.argmin(tot_dist)
        dist_2 = tot_dist[idx]
        src_2, det_2 = ref_lab[det_idx[idx]], ref_lab[src_idx[idx]]
        if dist_1 < dist_2:
            new_dist, src_use, det_use = dist_1, src_1, det_1
        else:
            new_dist, src_use, det_use = dist_2, det_2, src_2
        warn(
            "No fOLD table entry for best matching source/detector pair "
            f"{src_name}/{det_name} (RMS distance to the channel positions "
            f"was {1000 * dist:0.1f} mm) for channel index {cidx}, "
            "using next smallest available "
            f"src/det pairing {src_use}/{det_use} (RMS distance "
            f"{1000 * new_dist:0.1f} mm). Consider setting your channel "
            "positions to standard 10-05 locations using raw.set_montage "
            "if your pair does show up in the tables.",
            module="mne_nirs",
            ignore_namespaces=("mne", "mne_nirs"),
        )
        tbl = fold_tbl.query("Source == @src_use and Detector == @det_use")
        tbl = tbl.copy()
        tbl["BestSource"] = src_name
        tbl["BestDetector"] = det_name
        tbl["BestMatchDistance"] = dist
        tbl["MatchDistance"] = new_dist
        assert len(tbl)
    else:
        tbl = tbl.copy()
        tbl["BestSource"] = src_name
        tbl["BestDetector"] = det_name
        tbl["BestMatchDistance"] = dist
        tbl["MatchDistance"] = dist

    tbl = tbl.copy()  # don't get warnings about setting values later
    return tbl



# Authors: Robert Luke <mail@robertluke.net>
#
# License: BSD (3-clause)

import warnings
from collections import defaultdict

import mne
import numpy as np
import pytest
from mne.datasets import testing
from mne.utils import catch_logging, check_version

import mne_nirs
from mne_nirs.experimental_design import make_first_level_design_matrix
from mne_nirs.experimental_design.tests.test_experimental_design import _load_dataset
from mne_nirs.statistics import run_glm
from mne_nirs.statistics.tests.test_glm_type import _get_glm_result
from mne_nirs.utils import glm_to_tidy
from mne_nirs.visualisation import plot_glm_surface_projection

testing_path = testing.data_path(download=False)
raw_path = str(testing_path) + "/NIRx/nirscout/nirx_15_2_recording_w_short"
subjects_dir = str(testing_path) + "/subjects"


def test_plot_nirs_source_detector_pyvista(requires_pyvista):
    raw = mne.io.read_raw_nirx(raw_path)

    with warnings.catch_warnings():
        # Ignore deprecation warning caused by
        # app.setAttribute(Qt.AA_UseHighDpiPixmaps) in mne-python
        warnings.filterwarnings("ignore", category=DeprecationWarning)
        mne_nirs.visualisation.plot_nirs_source_detector(
            np.random.randn(len(raw.ch_names)),
            raw.info,
            show_axes=True,
            subject="fsaverage",
            trans="fsaverage",
            surfaces=["white"],
            fnirs=False,
            subjects_dir=subjects_dir,
            verbose=True,
        )

        mne_nirs.visualisation.plot_nirs_source_detector(
            np.abs(np.random.randn(len(raw.ch_names))) + 5,
            raw.info,
            show_axes=True,
            subject="fsaverage",
            trans="fsaverage",
            surfaces=["white"],
            fnirs=False,
            subjects_dir=subjects_dir,
            verbose=True,
        )


def test_run_plot_GLM_topo():
    raw_intensity = _load_dataset()
    raw_intensity.crop(450, 600)  # Keep the test fast

    design_matrix = make_first_level_design_matrix(
        raw_intensity, drift_order=1, drift_model="polynomial"
    )
    raw_od = mne.preprocessing.nirs.optical_density(raw_intensity)
    raw_haemo = mne.preprocessing.nirs.beer_lambert_law(raw_od, ppf=0.1)
    glm_estimates = run_glm(raw_haemo, design_matrix)
    fig = glm_estimates.plot_topo()
    # 5 conditions (A,B,C,Drift,Constant) * two chroma + 2xcolorbar
    assert len(fig.axes) == 12

    # Two conditions * two chroma + 2 x colorbar
    fig = glm_estimates.plot_topo(conditions=["A", "B"])
    assert len(fig.axes) == 6

    # One conditions * two chroma + 2 x colorbar
    fig = glm_estimates.plot_topo(conditions=["A"])
    assert len(fig.axes) == 4


def test_run_plot_GLM_contrast_topo():
    raw_intensity = _load_dataset()
    raw_intensity.crop(450, 600)  # Keep the test fast

    design_matrix = make_first_level_design_matrix(
        raw_intensity, drift_order=1, drift_model="polynomial"
    )
    raw_od = mne.preprocessing.nirs.optical_density(raw_intensity)
    raw_haemo = mne.preprocessing.nirs.beer_lambert_law(raw_od, ppf=0.1)
    glm_est = run_glm(raw_haemo, design_matrix)
    contrast_matrix = np.eye(design_matrix.shape[1])
    basic_conts = dict(
        [(column, contrast_matrix[i]) for i, column in enumerate(design_matrix.columns)]
    )
    contrast_LvR = basic_conts["A"] - basic_conts["B"]
    contrast = glm_est.compute_contrast(contrast_LvR)
    fig = contrast.plot_topo()
    assert len(fig.axes) == 3


def test_run_plot_GLM_contrast_topo_single_chroma():
    raw_intensity = _load_dataset()
    raw_intensity.crop(450, 600)  # Keep the test fast

    design_matrix = make_first_level_design_matrix(
        raw_intensity, drift_order=1, drift_model="polynomial"
    )
    raw_od = mne.preprocessing.nirs.optical_density(raw_intensity)
    raw_haemo = mne.preprocessing.nirs.beer_lambert_law(raw_od, ppf=0.1)
    raw_haemo = raw_haemo.pick(picks="hbo")
    glm_est = run_glm(raw_haemo, design_matrix)
    contrast_matrix = np.eye(design_matrix.shape[1])
    basic_conts = dict(
        [(column, contrast_matrix[i]) for i, column in enumerate(design_matrix.columns)]
    )
    contrast_LvR = basic_conts["A"] - basic_conts["B"]
    contrast = glm_est.compute_contrast(contrast_LvR)
    fig = contrast.plot_topo()
    assert len(fig.axes) == 2


def test_fig_from_axes():
    from mne_nirs.visualisation._plot_GLM_topo import _get_fig_from_axes

    with pytest.raises(RuntimeError, match="Unable to extract figure"):
        _get_fig_from_axes([1, 2, 3])


# surface arg
def test_run_plot_GLM_projection(requires_pyvista):
    raw_intensity = _load_dataset()
    raw_intensity.crop(450, 600)  # Keep the test fast

    design_matrix = make_first_level_design_matrix(
        raw_intensity, drift_order=1, drift_model="polynomial"
    )
    raw_od = mne.preprocessing.nirs.optical_density(raw_intensity)
    raw_haemo = mne.preprocessing.nirs.beer_lambert_law(raw_od, ppf=0.1)
    glm_estimates = run_glm(raw_haemo, design_matrix)
    df = glm_to_tidy(raw_haemo, glm_estimates.data, design_matrix)
    df = df.query("Chroma in 'hbo'")
    df = df.query("Condition in 'A'")

    with warnings.catch_warnings():
        # Ignore deprecation warning caused by
        # app.setAttribute(Qt.AA_UseHighDpiPixmaps) in mne-python
        warnings.filterwarnings("ignore", category=DeprecationWarning)
        brain = plot_glm_surface_projection(
            raw_haemo.copy().pick("hbo"),
            df,
            clim="auto",
            view="dorsal",
            colorbar=True,
            size=(800, 700),
            value="theta",
            surface="white",
            subjects_dir=subjects_dir,
        )
        assert isinstance(brain, mne.viz._brain.Brain)


@pytest.mark.parametrize(
    "fname_raw, to_1020, ch_names",
    [
        (raw_path, False, None),
        (raw_path, True, "numbered"),
        (raw_path, True, defaultdict(lambda: "")),
    ],
)
def test_plot_3d_montage(requires_pyvista, fname_raw, to_1020, ch_names):
    raw = mne.io.read_raw_nirx(fname_raw)
    if to_1020:
        need = set(
            sum((ch_name.split()[0].split("_") for ch_name in raw.ch_names), list())
        )
        mon = mne.channels.make_standard_montage("standard_1020")
        mon.rename_channels({h: n for h, n in zip(mon.ch_names, need)})
        raw.set_montage(mon)
    n_labels = len(raw.ch_names) // 2
    view_map = {
        "left-lat": np.arange(1, n_labels // 2),
        "caudal": np.arange(n_labels // 2, n_labels + 1),
    }
    # We use "sample" here even though it's wrong so that we can have a head
    # surface
    with catch_logging() as log, warnings.catch_warnings():
        # Ignore deprecation warning caused by
        # app.setAttribute(Qt.AA_UseHighDpiPixmaps) in mne-python
        warnings.filterwarnings("ignore", category=DeprecationWarning)
        mne_nirs.viz.plot_3d_montage(
            raw.info,
            view_map,
            subject="sample",
            surface="white",
            subjects_dir=subjects_dir,
            ch_names=ch_names,
            verbose=True,
        )
    log = log.getvalue().lower()
    if to_1020:
        assert "automatically mapped" in log
    else:
        assert "could not" in log


# surface arg
@pytest.mark.skipif(not check_version("mne", "1.0"), reason="Needs MNE-Python 1.0")
def test_glm_surface_projection(requires_pyvista):
    res = _get_glm_result(tmax=2974, tmin=0)
    with warnings.catch_warnings():
        # Ignore deprecation warning caused by
        # app.setAttribute(Qt.AA_UseHighDpiPixmaps) in mne-python
        warnings.filterwarnings("ignore", category=DeprecationWarning)
        res.surface_projection(
            condition="e3p0", view="dorsal", surface="white", subjects_dir=subjects_dir
        )
    with pytest.raises(KeyError, match="not found in conditions"):
        res.surface_projection(condition="foo")



# Authors: Robert Luke <code@robertluke.net>
# License: BSD (3-clause)

import os.path as op

import mne_nirs


def test_dataset_snirf_aux():
    datapath = mne_nirs.datasets.snirf_with_aux.data_path()
    assert op.isfile(datapath)
    assert "2022-08-05_002.snirf" in str(datapath)



# Authors: Robert Luke <mail@robertluke.net>
#
# License: BSD (3-clause)

import os

import mne
import numpy as np
import pytest

from mne_nirs.channels import get_long_channels, get_short_channels
from mne_nirs.signal_enhancement import short_channel_regression


def _load_dataset():
    """Load data and tidy it a bit"""
    fnirs_data_folder = mne.datasets.fnirs_motor.data_path()
    fnirs_raw_dir = os.path.join(fnirs_data_folder, "Participant-1")
    raw_intensity = mne.io.read_raw_nirx(fnirs_raw_dir, verbose=True).load_data()

    raw_intensity.crop(0, raw_intensity.annotations.onset[-1])

    new_des = [des for des in raw_intensity.annotations.description]
    new_des = ["A" if x == "1.0" else x for x in new_des]
    new_des = ["B" if x == "2.0" else x for x in new_des]
    new_des = ["C" if x == "3.0" else x for x in new_des]
    annot = mne.Annotations(
        raw_intensity.annotations.onset, raw_intensity.annotations.duration, new_des
    )
    raw_intensity.set_annotations(annot)

    assert "fnirs_cw_amplitude" in raw_intensity
    assert len(np.unique(raw_intensity.annotations.description)) == 4

    return raw_intensity


def test_short():
    raw_intensity = _load_dataset()

    with pytest.raises(RuntimeError, match="must be optical density"):
        _ = short_channel_regression(raw_intensity)

    raw_od = mne.preprocessing.nirs.optical_density(raw_intensity)

    raw_od_corrected = short_channel_regression(raw_od)

    assert "fnirs_od" in raw_od_corrected

    with pytest.raises(RuntimeError, match="long channels present"):
        short_channel_regression(get_short_channels(raw_od))
    with pytest.raises(RuntimeError, match="short channels present"):
        short_channel_regression(get_long_channels(raw_od))



# Authors: Robert Luke <mail@robertluke.net>
#
# License: BSD (3-clause)

import numpy as np
from mne.utils import warn


def glm_region_of_interest(glm, group_by, cond_idx, cond_name, weighted=True):
    """
    Calculate statistics for region of interest.

    Parameters
    ----------
    glm : dict
        Need to write.
    group_by : dict
        Specifies which channels are aggregated into a single ROI.
        The dict key will be used as the ROI label and the dict
        values must be lists of picks (either channel names or integer indices
        of ``epochs.ch_names``). For example::

            group_by=dict(Left_ROI=[1, 2, 3, 4], Right_ROI=[5, 6, 7, 8])

        Note that within a dict entry all channels must have the same type.
    cond_idx : int
        Index of condition of interest.
    cond_name : str
        Name to be used for condition.
    weighted : Bool
        Should channels be weighted by inverse of standard error (True).

    Returns
    -------
    stats : DataFrame
        Statistics for each ROI.
    """
    warn(
        '"glm_region_of_interest" has been deprecated in favor of the more '
        "comprehensive GLM class and will be removed in v1.0.0. "
        'Use the RegressionResults class "region_of_interest_dataframe()" '
        "method instead.",
        DeprecationWarning,
    )

    return _glm_region_of_interest(
        glm, group_by, cond_idx, cond_name, weighted=weighted
    )


def _glm_region_of_interest(stats, group_by, cond_idx, cond_name, weighted=True):
    """
    Calculate statistics for region of interest.

    Parameters
    ----------
    stats : dict
        Need to write.
    group_by : dict
        Specifies which channels are aggregated into a single ROI.
        The dict key will be used as the ROI label and the dict
        values must be lists of picks (either channel names or integer indices
        of ``epochs.ch_names``). For example::

            group_by=dict(Left_ROI=[1, 2, 3, 4], Right_ROI=[5, 6, 7, 8])

        Note that within a dict entry all channels must have the same type.
    cond_idx : int
        Index of condition of interest.
    cond_name : str
        Name to be used for condition.
    weighted : Bool | dict
        Weighting to be applied to each channel in the ROI computation.
        If False, then all channels will be weighted equally.
        If True, channels will be weighted by the inverse of
        the standard error of the GLM fit.
        For manual specification of the channel weighting a dictionary
        can be provided.
        If a dictionary is provided, the keys and length of lists must
        match the ``group_by`` parameters.
        The weights will be scaled internally to sum to 1.

    Returns
    -------
    stats : DataFrame
        Statistics for each ROI.
    """
    import pandas as pd
    from scipy import stats as ss

    df = pd.DataFrame()

    ch_names = np.array(list(stats.keys()))
    chromas = np.array([name[-3:] for name in ch_names])

    for region in group_by:
        if isinstance(weighted, dict):
            weights_region = weighted[region]

        roi_name = region
        picks = group_by[region]

        for chroma in np.unique(chromas[picks]):
            chroma_idxs = np.where([c == chroma for c in chromas[picks]])[0]
            chroma_picks = [picks[ci] for ci in chroma_idxs]

            thetas = list()
            ses = list()
            dfe = 0

            for pick in chroma_picks:
                thetas.append(1e6 * (stats[ch_names[pick]].theta)[cond_idx][0])
                ses.append(_se(stats[ch_names[pick]])[cond_idx])
                dfe = stats[ch_names[pick]].df_model

            # Apply weighting by standard error or custom values
            if weighted is True:
                weights = 1.0 / np.asarray(ses)
            elif weighted is False:
                weights = np.ones((len(ses), 1))
            elif isinstance(weighted, dict):
                weights = [float(weights_region[ci]) for ci in chroma_idxs]
                weights = np.asarray(weights).reshape((len(ses), 1))
            else:
                raise ValueError("Weighted parameter is not bool or dict")
            # Ensure weights sum to one
            weights /= np.sum(weights)

            theta = np.sum(thetas * weights)
            s = np.sum(ses * weights)

            t = theta / s

            p = 2 * ss.t.cdf(-1.0 * np.abs(t), df=dfe)

            this_df = pd.DataFrame(
                {
                    "ROI": roi_name,
                    "Condition": cond_name,
                    "Chroma": chroma,
                    "theta": theta / 1.0e6,
                    "se": s,
                    "t": t,
                    "dfe": dfe,
                    "p": p,
                },
                index=[0],
            )
            df = pd.concat([df, this_df], ignore_index=True)

    df.reset_index(inplace=True, drop=True)

    return df


def _se(d):
    return np.sqrt(np.diag(d.vcov())) * 1e6



# Authors: Robert Luke <mail@robertluke.net>
#
# License: BSD (3-clause)

import numpy as np
import pandas as pd
import pytest
import statsmodels.formula.api as smf
from mne.utils import check_version
from numpy.testing import assert_allclose

from mne_nirs.experimental_design import make_first_level_design_matrix
from mne_nirs.simulation import simulate_nirs_raw
from mne_nirs.statistics import run_glm, statsmodels_to_results


@pytest.mark.skipif(not check_version("lxml"), reason="Requires lxml")
@pytest.mark.parametrize("func", ("mixedlm", "ols", "rlm"))
@pytest.mark.filterwarnings("ignore:.*optimization.*:")
@pytest.mark.filterwarnings("ignore:.*unknown kwargs.*:")
@pytest.mark.filterwarnings("ignore:.*on the boundary.*:")
@pytest.mark.filterwarnings("ignore:.*The Hessian matrix at the estimated.*:")
def test_statsmodel_to_df(func):
    func = getattr(smf, func)
    np.random.seed(0)

    amplitude = 1.432

    df_cha = pd.DataFrame()
    for n in range(5):
        raw = simulate_nirs_raw(
            sfreq=3.0,
            amplitude=amplitude,
            sig_dur=300.0,
            stim_dur=5.0,
            isi_min=15.0,
            isi_max=45.0,
        )
        raw._data += np.random.normal(0, np.sqrt(1e-12), raw._data.shape)
        design_matrix = make_first_level_design_matrix(raw, stim_dur=5.0)
        glm_est = run_glm(raw, design_matrix)
        with pytest.warns(RuntimeWarning, match="Non standard source detect"):
            cha = glm_est.to_dataframe()
        cha["ID"] = "%02d" % n
        df_cha = pd.concat([df_cha, cha], ignore_index=True)
    df_cha["theta"] = df_cha["theta"] * 1.0e6
    roi_model = func("theta ~ -1 + Condition", df_cha, groups=df_cha["ID"]).fit()
    df = statsmodels_to_results(roi_model)
    assert isinstance(df, pd.DataFrame)
    assert_allclose(df["Coef."]["Condition[A]"], amplitude, rtol=0.1)
    assert df["Significant"]["Condition[A]"]
    assert df.shape == (8, 8)

    roi_model = smf.rlm("theta ~ -1 + Condition", df_cha, groups=df_cha["ID"]).fit()
    df = statsmodels_to_results(roi_model)
    assert isinstance(df, pd.DataFrame)
    assert_allclose(df["Coef."]["Condition[A]"], amplitude, rtol=0.1)
    assert df["Significant"]["Condition[A]"]
    assert df.shape == (8, 8)



# Authors: Robert Luke <mail@robertluke.net>
#
# License: BSD (3-clause)

import re

import numpy as np
from mne import pick_types
from mne.io import BaseRaw
from mne.utils import _validate_type


def list_sources(raw):
    """
    List all the sources in the fNIRS montage.

    Parameters
    ----------
    raw : instance of Raw
        Raw instance containing fNIRS data.

    Returns
    -------
    sources : list
        Unique list of all sources in ascending order.
    """
    _validate_type(raw, BaseRaw, "raw")

    picks = pick_types(raw.info, meg=False, eeg=False, fnirs=True, exclude=[])
    if not len(picks):
        raise RuntimeError("Listing source is for fNIRS signals only.")

    sources = list()
    ch_names = raw.ch_names
    for pick in picks:
        x = re.search(r"S(\d+)_D(\d+)", ch_names[pick])
        sources.append(int(x.group(1)))

    sources = np.unique(sources)

    return sources


def list_detectors(raw):
    """
    List all the detectors in the fNIRS montage.

    Parameters
    ----------
    raw : instance of Raw
        Raw instance containing fNIRS data.

    Returns
    -------
    sources : list
        Unique list of all detectors in ascending order.
    """
    _validate_type(raw, BaseRaw, "raw")

    picks = pick_types(raw.info, meg=False, eeg=False, fnirs=True, exclude=[])
    if not len(picks):
        raise RuntimeError("Listing source is for fNIRS signals only.")

    detectors = list()
    ch_names = raw.ch_names
    for pick in picks:
        x = re.search(r"S(\d+)_D(\d+)", ch_names[pick])
        detectors.append(int(x.group(2)))

    detectors = np.unique(detectors)

    return detectors


def drop_sources(raw, sources):
    """
    Drop sources.

    Parameters
    ----------
    raw : instance of Raw
        Raw instance containing fNIRS data.
    sources : iterable or str
        Iterable (e.g. list) of source(s) or source to remove.

    Returns
    -------
    inst : instance of Raw
        The modified instance.
    """
    if isinstance(sources, int):
        sources = [sources]

    try:
        all_str = all([isinstance(src, int) for src in sources])
    except TypeError:
        raise ValueError(
            f"'ch_names' must be iterable, got type {type(sources)} ({sources})."
        )

    if not all_str:
        raise ValueError(
            "Each element in 'ch_names' must be int, got "
            f"{[type(ch) for ch in sources]}."
        )

    keeps = np.ones(len(raw.ch_names))
    for src in sources:
        template = f"S{src}_"
        for idx in range(len(raw.ch_names)):
            if template in raw.ch_names[idx]:
                keeps[idx] = 0

    keeps = np.where(keeps)[0]

    return raw.pick(keeps)


def drop_detectors(raw, detectors):
    """
    Drop detectors.

    Parameters
    ----------
    raw : instance of Raw
        Raw instance containing fNIRS data.
    detectors : iterable or str
        Iterable (e.g. list) of detector(s) or detector to remove.

    Returns
    -------
    inst : instance of Raw
        The modified instance.
    """
    if isinstance(detectors, int):
        detectors = [detectors]

    try:
        all_str = all([isinstance(det, int) for det in detectors])
    except TypeError:
        raise ValueError(
            f"'ch_names' must be iterable, got type {type(detectors)} ({detectors})."
        )

    if not all_str:
        raise ValueError(
            "Each element in 'ch_names' must be int, got "
            f"{[type(det) for det in detectors]}."
        )

    keeps = np.ones(len(raw.ch_names))
    for det in detectors:
        template = f"_D{det} "
        for idx in range(len(raw.ch_names)):
            if template in raw.ch_names[idx]:
                keeps[idx] = 0

    keeps = np.where(keeps)[0]

    return raw.pick(keeps)


def pick_sources(raw, sources):
    """
    Pick sources.

    Parameters
    ----------
    raw : instance of Raw
        Raw instance containing fNIRS data.
    sources : iterable or str
        Iterable (e.g. list) of source(s) or source to keep.

    Returns
    -------
    inst : instance of Raw
        The modified instance.
    """
    if isinstance(sources, int):
        sources = [sources]

    try:
        all_str = all([isinstance(src, int) for src in sources])
    except TypeError:
        raise ValueError(
            f"'ch_names' must be iterable, got type {type(sources)} ({sources})."
        )

    if not all_str:
        raise ValueError(
            "Each element in 'ch_names' must be int, got "
            f"{[type(ch) for ch in sources]}."
        )

    keeps = np.zeros(len(raw.ch_names))
    for src in sources:
        template = f"S{src}_"
        for idx in range(len(raw.ch_names)):
            if template in raw.ch_names[idx]:
                keeps[idx] = 1

    keeps = np.where(keeps)[0]

    return raw.pick(keeps)


def pick_detectors(raw, detectors):
    """
    Pick detectors.

    Parameters
    ----------
    raw : instance of Raw
        Raw instance containing fNIRS data.
    detectors : iterable or str
        Iterable (e.g. list) of detector(s) or detector to keep.

    Returns
    -------
    inst : instance of Raw
        The modified instance.
    """
    if isinstance(detectors, int):
        detectors = [detectors]

    try:
        all_str = all([isinstance(det, int) for det in detectors])
    except TypeError:
        raise ValueError(
            f"'ch_names' must be iterable, got type {type(detectors)} ({detectors})."
        )

    if not all_str:
        raise ValueError(
            "Each element in 'ch_names' must be int, got "
            f"{[type(det) for det in detectors]}."
        )

    keeps = np.zeros(len(raw.ch_names))
    for det in detectors:
        template = f"_D{det} "
        for idx in range(len(raw.ch_names)):
            if template in raw.ch_names[idx]:
                keeps[idx] = 1

    keeps = np.where(keeps)[0]

    return raw.pick(keeps)



# Authors: Robert Luke <mail@robertluke.net>
#
# License: BSD (3-clause)

import inspect
from copy import deepcopy

import matplotlib as mpl
import matplotlib.pyplot as plt
import numpy as np
from mne import Info, pick_info
from mne.channels.layout import _merge_ch_data
from mne.io.pick import _picks_to_idx
from mne.utils import warn
from mne.viz import plot_topomap
from mpl_toolkits.axes_grid1.axes_divider import make_axes_locatable


def _plot_glm_topo(
    inst,
    glm_estimates,
    design_matrix,
    *,
    requested_conditions=None,
    axes=None,
    vlim=None,
    vmin=None,
    vmax=None,
    colorbar=True,
    figsize=(12, 7),
    sphere=None,
    **kwargs,
):
    info = deepcopy(inst if isinstance(inst, Info) else inst.info)

    if not (info.ch_names == list(glm_estimates.keys())):
        if len(info.ch_names) < len(list(glm_estimates.keys())):
            warn("Reducing GLM results to match MNE data")
            glm_estimates = {a: glm_estimates[a] for a in info.ch_names}
        else:
            raise RuntimeError(
                "MNE data structure does not match regression "
                f"results. Raw = {len(info.ch_names)}. "
                f"GLM = {len(list(glm_estimates.keys()))}"
            )

    estimates = np.zeros((len(glm_estimates), len(design_matrix.columns)))

    for idx, name in enumerate(glm_estimates.keys()):
        estimates[idx, :] = glm_estimates[name].theta.T

    types = np.unique(info.get_channel_types())

    if requested_conditions is None:
        requested_conditions = design_matrix.columns
    requested_conditions = [
        x for x in design_matrix.columns if x in requested_conditions
    ]

    # Plotting setup
    if axes is None:
        fig, axes = plt.subplots(
            nrows=len(types), ncols=len(requested_conditions), figsize=figsize
        )

    estimates = estimates[:, [c in requested_conditions for c in design_matrix.columns]]

    estimates = estimates * 1e6
    design_matrix = design_matrix[requested_conditions]
    vlim, vlim_kwargs = _handle_vlim(vlim, vmin, vmax, estimates)
    del vmin, vmax
    cmap = mpl.cm.RdBu_r
    norm = mpl.colors.Normalize(vmin=vlim[0], vmax=vlim[1])

    for t_idx, t in enumerate(types):
        estmrg, pos, chs, sphere = _handle_overlaps(info, t, sphere, estimates)

        for idx, label in enumerate(design_matrix.columns):
            if label in requested_conditions:
                # Deal with case when only a single
                # chroma or condition is available
                if (len(requested_conditions) == 1) & (len(types) == 1):
                    ax = axes
                elif (len(requested_conditions) == 1) & (len(types) > 1):
                    ax = axes[t_idx]
                elif (len(requested_conditions) > 1) & (len(types) == 1):
                    ax = axes[idx]
                else:
                    ax = axes[t_idx, idx]

                plot_topomap(
                    estmrg[:, idx],
                    pos,
                    extrapolate="local",
                    names=chs,
                    cmap=cmap,
                    axes=ax,
                    show=False,
                    sphere=sphere,
                    **vlim_kwargs,
                    **kwargs,
                )
                ax.set_title(label)

        if colorbar:
            ax1_divider = make_axes_locatable(ax)
            cax1 = ax1_divider.append_axes("right", size="7%", pad="2%")
            cbar = mpl.colorbar.ColorbarBase(
                cax1, cmap=cmap, norm=norm, orientation="vertical"
            )
            cbar.set_label("Haemoglobin (uM)", rotation=270)

    return _get_fig_from_axes(axes)


def _plot_glm_contrast_topo(inst, contrast, figsize=(12, 7), sphere=None, **kwargs):
    info = deepcopy(inst if isinstance(inst, Info) else inst.info)

    # Extract types. One subplot is created per type (hbo/hbr)
    types = np.unique(info.get_channel_types())

    # Extract values to plot and rescale to uM
    estimates = contrast.effect
    if estimates.ndim == 2:  # old nilearn
        assert estimates.shape[0] == 1
        estimates = estimates[0]
    estimates = estimates * 1e6

    # Create subplots for figures
    fig, axes = plt.subplots(nrows=1, ncols=len(types), figsize=figsize)
    # Create limits for colorbar
    vlim, vlim_kwargs = _handle_vlim((None, None), None, None, estimates)
    cmap = mpl.cm.RdBu_r
    norm = mpl.colors.Normalize(vmin=vlim[0], vmax=vlim[1])

    for t_idx, t in enumerate(types):
        estmrg, pos, chs, sphere = _handle_overlaps(info, t, sphere, estimates)

        # Deal with case when only a single chroma is available
        if len(types) == 1:
            ax = axes
        else:
            ax = axes[t_idx]

        # Plot the topomap
        plot_topomap(
            estmrg,
            pos,
            extrapolate="local",
            names=chs,
            cmap=cmap,
            axes=ax,
            show=False,
            sphere=sphere,
            **vlim_kwargs,
            **kwargs,
        )
        # Sets axes title
        if t == "hbo":
            ax.set_title("Oxyhaemoglobin")
        elif t == "hbr":
            ax.set_title("Deoxyhaemoglobin")
        else:
            ax.set_title(t)

    # Create a single colorbar for all types based on limits above
    ax1_divider = make_axes_locatable(ax)
    cax1 = ax1_divider.append_axes("right", size="7%", pad="2%")
    cbar = mpl.colorbar.ColorbarBase(cax1, cmap=cmap, norm=norm, orientation="vertical")
    cbar.set_label("Contrast Effect", rotation=270)

    return fig


def plot_glm_group_topo(
    inst,
    statsmodel_df,
    value="Coef.",
    axes=None,
    threshold=False,
    *,
    vlim=(None, None),
    vmin=None,
    vmax=None,
    cmap=None,
    sensors=True,
    res=64,
    sphere=None,
    colorbar=True,
    names=False,
    show_names=None,
    extrapolate="local",
    image_interp="cubic",
):
    """
    Plot topomap of NIRS group level GLM results.

    Parameters
    ----------
    inst : instance of Info or Raw
        Raw data or info structure used to generate the GLM results.
    statsmodel_df : DataFrame
        Dataframe created from a statsmodel summary.
    value : String
        Which column in the `statsmodel_df` to use in the topo map.
    axes : instance of Axes | None
        The axes to plot to. If None, the current axes will be used.
    threshold : Bool
        If threshold is true, all values with P>|z| greater than 0.05 will
        be set to zero.
    vlim : tuple of length 2
        Colormap limits to use. If a :class:`tuple` of floats, specifies the
        lower and upper bounds of the colormap (in that order); providing
        ``None`` for either entry will set the corresponding boundary at the
        min/max of the data (separately for each topomap).
    vmin : float | None
        Deprecated, use 'vlim' instead.
    vmax : float | None
        Deprecated, use 'vlim' instead.
    cmap : matplotlib colormap | None
        Colormap to use. If None, 'Reds' is used for all positive data,
        otherwise defaults to 'RdBu_r'.
    sensors : bool | str
        Add markers for sensor locations to the plot. Accepts matplotlib plot
        format string (e.g., 'r+' for red plusses). If True (default), circles
        will be used.
    res : int
        The resolution of the topomap image (n pixels along each side).
    sphere : numbers
        As specified in mne.
    colorbar : bool
        Should a colorbar be plotted.
    names : list of str
        The channel names to display.
    show_names : bool
        Deprecated, use ``names`` instead.
    extrapolate : str
        Type of extrapolation for image.
    image_interp : str
        Type of interpolation for image.

    Returns
    -------
    fig : figure
        Figure with topographic representation of statsmodel_df value.
    """
    info = deepcopy(inst if isinstance(inst, Info) else inst.info)
    if show_names is not None:
        names = show_names
        warn(
            "show_names is deprecated and will be removed in the next "
            "release, use names instead",
            FutureWarning,
        )
    del show_names

    # Check that the channels in two inputs match
    if not (info.ch_names == list(statsmodel_df["ch_name"].values)):
        if len(info.ch_names) < len(list(statsmodel_df["ch_name"].values)):
            print("Reducing GLM results to match MNE data")
            statsmodel_df["Keep"] = [
                g in info.ch_names for g in statsmodel_df["ch_name"]
            ]
            statsmodel_df = statsmodel_df.query("Keep == True")
        else:
            warn("MNE data structure does not match regression results")
    statsmodel_df = statsmodel_df.set_index("ch_name")
    statsmodel_df = statsmodel_df.reindex(info.ch_names)

    # Extract estimate of interest to plot
    estimates = statsmodel_df[value].values

    if threshold:
        p = statsmodel_df["P>|z|"].values
        t = p > 0.05
        estimates[t] = 0.0

    assert len(np.unique(statsmodel_df["Chroma"])) == 1, "Only one Chroma allowed"

    if "Condition" in statsmodel_df.columns:
        assert len(np.unique(statsmodel_df["Condition"])) == 1, (
            "Only one condition allowed"
        )
        c = np.unique(statsmodel_df["Condition"])[0]
    else:
        c = "Contrast"

    t = np.unique(statsmodel_df["Chroma"])[0]

    # Plotting setup
    if axes is None:
        fig, axes = plt.subplots(nrows=1, ncols=1, figsize=(12, 7))
    # Set limits of topomap and colors
    vlim, vlim_kwargs = _handle_vlim(vlim, vmin, vmax, estimates)
    del vmin, vmax
    if cmap is None:
        cmap = mpl.cm.RdBu_r
    norm = mpl.colors.Normalize(vmin=vlim[0], vmax=vlim[1])

    estmrg, pos, chs, sphere = _handle_overlaps(info, t, sphere, estimates)
    if "names" in inspect.signature(plot_topomap).parameters:
        names_kwarg = dict(names=chs if names else [""] * len(chs))
    else:
        names_kwarg = dict(show_names=names, names=chs)

    plot_topomap(
        estmrg,
        pos,
        extrapolate=extrapolate,
        image_interp=image_interp,
        cmap=cmap,
        axes=axes,
        sensors=sensors,
        res=res,
        show=False,
        sphere=sphere,
        **vlim_kwargs,
        **names_kwarg,
    )
    axes.set_title(c)

    if colorbar:
        ax1_divider = make_axes_locatable(axes)
        cax1 = ax1_divider.append_axes("right", size="7%", pad="2%")
        cbar = mpl.colorbar.ColorbarBase(
            cax1, cmap=cmap, norm=norm, orientation="vertical"
        )
        cbar.set_label(value, rotation=270)

    return axes


def _handle_overlaps(info, t, sphere, estimates):
    """Prepare for topomap including merging channels."""
    from mne.viz.topomap import _prepare_topomap_plot

    picks = _picks_to_idx(info, t, exclude=[], allow_empty=True)
    info_subset = pick_info(info, picks)
    (
        _,
        pos,
        merge_channels,
        ch_names,
        ch_type,
        sphere,
        clip_origin,
    ) = _prepare_topomap_plot(info_subset, t, sphere=sphere)
    estmrg, ch_names = _merge_ch_data(estimates.copy()[picks], t, ch_names)
    return estmrg, pos, ch_names, sphere


def _get_fig_from_axes(ax):
    if isinstance(ax, mpl.axes.SubplotBase):
        return ax.figure
    elif type(ax) is np.ndarray:
        return _get_fig_from_axes(ax[0])
    else:
        raise RuntimeError(f"Unable to extract figure from {ax}")


def _handle_vlim(vlim, vmin, vmax, estimates):
    if vmin is not None or vmax is not None:
        warn(
            "vmin and vmax are deprecated and will be removed in the next "
            "release, please use vlim instead",
            FutureWarning,
        )
        vlim = (vmin, vmax)
    else:
        vmin, vmax = vlim
    if vmax is None:
        vmax = np.max(np.abs(estimates))
    if vmin is None:
        vmin = vmax * -1.0
    vlim = tuple(vlim)
    if "vlim" in inspect.signature(plot_topomap).parameters:
        kwargs = dict(vlim=(vmin, vmax))
    else:
        kwargs = dict(vmin=vmin, vmax=vmax)
    return vlim, kwargs



# Authors: Robert Luke <mail@robertluke.net>
#
# License: BSD (3-clause)

import numpy as np
from mne import Annotations, create_info
from mne.io import RawArray


def simulate_nirs_raw(
    sfreq=3.0,
    amplitude=1.0,
    annot_desc="A",
    sig_dur=300.0,
    stim_dur=5.0,
    isi_min=15.0,
    isi_max=45.0,
    ch_name="Simulated",
    hrf_model="glover",
):
    """
    Create simulated fNIRS data.

    The returned data is of type `hbo`.
    One or more conditions can be simulated.
    To simulate multiple conditions pass in a description and amplitude
    for each
    `amplitude=[0., 2., 4.], annot_desc=['Control', 'Cond_A', 'Cond_B']`.

    Parameters
    ----------
    sfreq : Number
        The sample rate.
    amplitude : Number, Array of numbers
        The amplitude of the signal to simulate in uM.
        Pass in an array to simulate multiple conditions.
    annot_desc : str, Array of str
        The name of the annotations for simulated amplitudes.
        Pass in an array to simulate multiple conditions,
        must be the same length as amplitude.
    sig_dur : Number
        The length of the boxcar signal to generate in seconds that will
        be convolved with the HRF.
    stim_dur : Number, Array of numbers
        The length of the stimulus to generate in seconds.
    isi_min : Number
        The minimum duration of the inter stimulus interval in seconds.
    isi_max : Number
        The maximum duration of the inter stimulus interval in seconds.
    ch_name : str
        Channel name to be used in returned raw instance.
    hrf_model : str
        Specifies the hemodynamic response function. See nilearn docs.

    Returns
    -------
    raw : instance of Raw
        The generated raw instance.
    """
    from nilearn.glm.first_level import make_first_level_design_matrix
    from pandas import DataFrame

    if not isinstance(amplitude, list):
        amplitude = [amplitude]
    if not isinstance(annot_desc, list):
        annot_desc = [annot_desc]
    if not isinstance(stim_dur, list):
        stim_dur = [stim_dur]

    frame_times = np.arange(sig_dur * sfreq) / sfreq

    assert len(amplitude) == len(annot_desc), (
        "Same number of amplitudes as annotations required."
    )
    assert len(amplitude) == len(stim_dur), (
        "Same number of amplitudes as durations required."
    )

    onset = 0.0
    onsets = []
    conditions = []
    durations = []
    while onset < sig_dur - 60:
        c_idx = np.random.randint(0, len(amplitude))
        onset += np.random.uniform(isi_min, isi_max) + stim_dur[c_idx]
        onsets.append(onset)
        conditions.append(annot_desc[c_idx])
        durations.append(stim_dur[c_idx])

    events = DataFrame(
        {"trial_type": conditions, "onset": onsets, "duration": durations}
    )

    dm = make_first_level_design_matrix(
        frame_times,
        events,
        hrf_model=hrf_model,
        drift_model="polynomial",
        drift_order=0,
    )
    dm = dm.drop(columns="constant")

    annotations = Annotations(onsets, durations, conditions)

    info = create_info(ch_names=[ch_name], sfreq=sfreq, ch_types=["hbo"])

    for idx, annot in enumerate(annot_desc):
        if annot in dm.columns:
            dm[annot] *= amplitude[idx]

    a = np.sum(dm.to_numpy(), axis=1) * 1.0e-6
    a = a.reshape(-1, 1).T

    raw = RawArray(a, info, verbose=False)
    raw.set_annotations(annotations, verbose="error")

    return raw



# Author: Robert Luke <mail@robertluke.net>
#
# License: BSD (3-clause)

from pathlib import Path
from shutil import copyfile

import mne
import numpy as np
import pandas as pd
import pytest
from mne.channels import make_standard_montage
from mne.channels.montage import transform_to_head
from mne.datasets.testing import data_path, requires_testing_data
from mne.io import read_fiducials, read_raw_nirx
from numpy.testing import assert_allclose

from mne_nirs.io import fold_landmark_specificity
from mne_nirs.io.fold import fold_channel_specificity
from mne_nirs.io.fold._fold import (
    _find_closest_standard_location,
    _generate_montage_locations,
    _read_fold_xls,
)

thisfile = Path(__file__).parent.resolve()
foldfile = thisfile / "data" / "example.xls"

# https://github.com/mne-tools/mne-testing-data/pull/72
fname_nirx_15_3_short = (
    Path(data_path(download=False)) / "NIRx" / "nirscout" / "nirx_15_3_recording"
)

pytest.importorskip("xlrd", "1.0")


@pytest.mark.parametrize("fold_files", (str, None, list))
def test_channel_specificity(monkeypatch, tmp_path, fold_files):
    raw = read_raw_nirx(fname_nirx_15_3_short, preload=True)
    raw.pick(range(2))
    kwargs = dict()
    n_want = 6
    if fold_files is list:
        kwargs = dict(fold_files=[foldfile])
    elif fold_files is str:
        kwargs = dict(fold_files=tmp_path)
        n_want *= 2
    else:
        assert fold_files is None
        monkeypatch.setenv("MNE_NIRS_FOLD_PATH", str(tmp_path))
        assert len(kwargs) == 0
        with pytest.raises(FileNotFoundError, match=r"fold_files\[0\] does.*"):
            fold_channel_specificity(raw)
        n_want *= 2
    copyfile(foldfile, tmp_path / "10-10.xls")
    copyfile(foldfile, tmp_path / "10-5.xls")
    res = fold_channel_specificity(raw, **kwargs)
    assert len(res) == 2
    assert res[0].shape == (n_want, 14)
    montage = make_standard_montage("standard_1005", head_size=0.09700884729534559)
    fids = read_fiducials(
        Path(mne.__file__).parent / "data" / "fsaverage" / "fsaverage-fiducials.fif"
    )[0]
    for f in fids:
        f["coord_frame"] = montage.dig[0]["coord_frame"]
    montage.dig[:3] = fids
    S, D = raw.ch_names[0].split()[0].split("_")
    assert S == "S1" and D == "D2"
    montage.rename_channels({"PO8": S, "P6": D})  # not in the tables!
    # taken from standard_1020.elc
    s_mri = np.array([55.6666, -97.6251, 2.7300]) / 1000.0
    d_mri = np.array([67.8877, -75.9043, 28.0910]) / 1000.0
    trans = mne.transforms._get_trans("fsaverage", "mri", "head")[0]
    ch_pos = montage.get_positions()["ch_pos"]
    assert_allclose(ch_pos[S], s_mri, atol=1e-6)
    assert_allclose(ch_pos[D], d_mri, atol=1e-6)
    raw.set_montage(montage)
    montage = transform_to_head(montage)
    s_head = mne.transforms.apply_trans(trans, s_mri)
    d_head = mne.transforms.apply_trans(trans, d_mri)
    assert_allclose(montage._get_ch_pos()["S1"], s_head, atol=1e-6)
    assert_allclose(montage._get_ch_pos()["D2"], d_head, atol=1e-6)
    for ch in raw.info["chs"]:
        assert_allclose(ch["loc"][3:6], s_head, atol=1e-6)
        assert_allclose(ch["loc"][6:9], d_head, atol=1e-6)
    res_1 = fold_channel_specificity(raw, **kwargs)[0]
    assert res_1.shape == (0, 14)
    # TODO: This is wrong, should be P08 not P08h, and distance should be 0 mm!
    with pytest.warns(RuntimeWarning, match=".*PO8h?/P6.*TP8/T8.*"):
        res_1 = fold_channel_specificity(raw, interpolate=True, **kwargs)[0]
    montage.rename_channels({S: D, D: S})  # reversed
    with pytest.warns(RuntimeWarning, match=".*PO8h?/P6.*TP8/T8.*"):
        res_2 = fold_channel_specificity(raw, interpolate=True, **kwargs)[0]
    # We should check the whole thing, but this is probably good enough
    assert (res_1["Specificity"] == res_2["Specificity"]).all()


def test_landmark_specificity():
    raw = read_raw_nirx(fname_nirx_15_3_short, preload=True)
    with pytest.warns(RuntimeWarning, match="No fOLD table entry"):
        res = fold_landmark_specificity(
            raw, "L Superior Frontal Gyrus", [foldfile], interpolate=True
        )
    assert len(res) == len(raw.ch_names)
    assert np.max(res) <= 100
    assert np.min(res) >= 0


def test_fold_workflow():
    # Read raw data
    raw = read_raw_nirx(fname_nirx_15_3_short, preload=True)
    reference_locations = _generate_montage_locations()
    channel_of_interest = raw.copy().pick(1)

    # Get source and detector labels
    source_locs = channel_of_interest.info["chs"][0]["loc"][3:6]
    source_label = _find_closest_standard_location(source_locs, reference_locations)
    assert source_label == "T7"

    detector_locs = channel_of_interest.info["chs"][0]["loc"][6:9]
    detector_label = _find_closest_standard_location(detector_locs, reference_locations)
    assert detector_label == "TP7"

    # Find correct fOLD elements
    tbl = _read_fold_xls(foldfile, atlas="Juelich")
    tbl = tbl.query("Source == @source_label").query("Detector == @detector_label")

    # Query region of interest
    specificity = tbl.query("Landmark == 'L Mid Orbital Gyrus'")["Specificity"]
    assert specificity.values == 12.34


def test_fold_reader():
    tbl = _read_fold_xls(foldfile, atlas="Juelich")
    assert isinstance(tbl, pd.DataFrame)
    assert tbl.shape == (11, 10)
    assert "L Superior Frontal Gyrus" in list(tbl["Landmark"])


@requires_testing_data
def test_label_finder():
    """Test locating labels."""
    raw = read_raw_nirx(fname_nirx_15_3_short, preload=True)
    reference_locations = _generate_montage_locations()

    # Test central head position source
    raw_tmp = raw.copy().pick(25)
    assert (
        _find_closest_standard_location(
            raw_tmp.info["chs"][0]["loc"][3:6], reference_locations
        )
        == "Cz"
    )

    # Test right auditory position detector
    raw_tmp = raw.copy().pick(4)
    assert (
        _find_closest_standard_location(
            raw_tmp.info["chs"][0]["loc"][6:9], reference_locations
        )
        == "T8"
    )

    # Test right auditory position source
    raw_tmp = raw.copy().pick(4)
    assert (
        _find_closest_standard_location(
            raw_tmp.info["chs"][0]["loc"][3:6], reference_locations
        )
        == "TP8"
    )

    # Test left auditory position source
    raw_tmp = raw.copy().pick(1)
    assert (
        _find_closest_standard_location(
            raw_tmp.info["chs"][0]["loc"][3:6], reference_locations
        )
        == "T7"
    )

    # Test left auditory position detector
    raw_tmp = raw.copy().pick(1)
    assert (
        _find_closest_standard_location(
            raw_tmp.info["chs"][0]["loc"][6:9], reference_locations
        )
        == "TP7"
    )



# Authors: Robert Luke <mail@robertluke.net>
#
# License: BSD (3-clause)

from ._short import get_short_channels, get_long_channels
from ._roi import picks_pair_to_idx
from ._channels import (
    list_sources,
    list_detectors,
    drop_sources,
    drop_detectors,
    pick_sources,
    pick_detectors,
)



# Author: Robert Luke <code@robertluke.net>
#
# License: BSD (3-clause)

import logging

import h5py
import numpy as np
from mne.io import Raw
from pandas import DataFrame
from scipy import interpolate


def read_snirf_aux_data(fname: str, raw: Raw):
    """Read auxiliary data from SNIRF file.

    Reads the auxiliary channel data (e.g. heart rate data,
    accelerometer, etc). The auxiliary data will be resampled
    to match the raw data.

    Parameters
    ----------
    fname : str
        Path to the SNIRF data file.
    raw : Raw
        Instance of raw snirf data as created by read_raw_snirf.

    Returns
    -------
    fname : str
        Path to the SNIRF data file.
    """
    with h5py.File(fname, "r") as dat:
        if "nirs" in dat:
            basename = "nirs"
        elif "nirs1" in dat:
            basename = "nirs1"
        else:
            raise RuntimeError("Data does not contain nirs field")

        all_keys = list(dat.get(basename).keys())
        aux_keys = [i for i in all_keys if i.startswith("aux")]
        aux_names = [_decode_name(dat.get(f"{basename}/{k}/name")) for k in aux_keys]
        logging.debug(f"Found auxiliary channels {aux_names}")

        d = {"times": raw.times}
        for idx, aux in enumerate(aux_keys):
            aux_data = np.array(dat.get(f"{basename}/{aux}/dataTimeSeries"))
            aux_time = np.array(dat.get(f"{basename}/{aux}/time"))
            aux_data_interp = interpolate.interp1d(
                aux_time, aux_data, axis=0, bounds_error=False, fill_value="extrapolate"
            )
            aux_data_matched_to_raw = aux_data_interp(raw.times)
            d[aux_names[idx]] = aux_data_matched_to_raw

        df = DataFrame(data=d)
        df = df.set_index("times")

    return df


def _decode_name(key):
    return np.array(key)[0].decode()



# Author: Eric Larson <larson.eric.d@gmail.com>
#
# License: BSD-3-Clause

import os
import warnings
from unittest import mock

import mne
import pytest

# most of this adapted from MNE-Python


def pytest_configure(config):
    """Configure pytest options."""
    # Markers
    for marker in ("examples",):
        config.addinivalue_line("markers", marker)
    for fixture in ("matplotlib_config", "close_all"):
        config.addinivalue_line("usefixtures", fixture)

    warning_lines = r"""
    error::
    ignore:.*np\.MachAr.*:DeprecationWarning
    ignore:.*sysconfig module is deprecated.*:DeprecationWarning
    ignore:.*nilearn.glm module is experimental.*:
    ignore:.*Using or importing the ABCs from.*:
    ignore:.*distutils Version classes are deprecated.*:
    ignore:.*LUTSIZE was deprecated in Matplotlib.*:
    ignore:.*pandas\.Int64Index is deprecated and will be removed.*:
    ignore:.*Setting non-standard config type.*:
    ignore:.*The MLE may be on the boundary.*:
    ignore:.*The Hessian matrix at the estimated parameter values.*:
    always:`np\..*is a deprecated alias for the builtin.*:DeprecationWarning
    ignore:.*data_path functions now return.*
    ignore:.*default value of `n_init`*
    ignore:.*get_cmap function will be deprecated`*
    ignore:.*The register_cmap function`*
    ignore:.*is not yet supported for.*in qdarkstyle.*:RuntimeWarning
    always::ResourceWarning
    ignore:_SixMetaPathImporter.find_spec\(\) not found.*:ImportWarning
    # Should probably fix these at some point...
    ignore:unclosed file.*:ResourceWarning
    # seaborn
    ignore:The register_cmap function.*:
    ignore:The get_cmap function.*:
    ignore:The figure layout has changed.*:UserWarning
    # H5py
    ignore:`product` is deprecated as of NumPy.*:DeprecationWarning
    # seaborn
    ignore:is_categorical_dtype is deprecated.*:FutureWarning
    ignore:use_inf_as_na option is deprecated.*:FutureWarning
    # nilearn
    ignore:The provided callable <function sum.*:FutureWarning
    ignore:The parameter "contrast_type" will be removed.*:DeprecationWarning
    # TODO: in an example (should fix eventually)
    ignore:The behavior of DataFrame concatenation.*:FutureWarning
    # mne-bids needs a release
    ignore:.*mne\.io\.pick.* is deprecated.*:FutureWarning
    # MESA
    ignore:Mesa version 10\.2\.4 is too old.*:RuntimeWarning
    # Pandas
    ignore:np\.find_common_type is deprecated.*:DeprecationWarning
    # statsmodels
    ignore:The numpy\.linalg\.linalg has been made private.*:DeprecationWarning
    # sklearn
    ignore:KMeans is known to have a memory leak on Windows with MKL.*:UserWarning
    ignore:.*The `disp` and `iprint` options of the L-BFGS-B.*:DeprecationWarning
    """  # noqa: E501
    for warning_line in warning_lines.split("\n"):
        warning_line = warning_line.strip()
        if warning_line and not warning_line.startswith("#"):
            config.addinivalue_line("filterwarnings", warning_line)


@pytest.fixture(scope="session")
def matplotlib_config():
    """Configure matplotlib for viz tests."""
    from matplotlib import cbook, use

    want = "agg"  # don't pop up windows
    with warnings.catch_warnings(record=True):  # ignore warning
        warnings.filterwarnings("ignore")
        use(want, force=True)
    import matplotlib.pyplot as plt

    assert plt.get_backend() == want
    # overwrite some params that can horribly slow down tests that
    # users might have changed locally (but should not otherwise affect
    # functionality)
    plt.ioff()
    plt.rcParams["figure.dpi"] = 100
    plt.rcParams["figure.max_open_warning"] = 100

    # Make sure that we always reraise exceptions in handlers
    orig = cbook.CallbackRegistry

    class CallbackRegistryReraise(orig):
        def __init__(self, exception_handler=None, signals=None):
            super().__init__(exception_handler)

    cbook.CallbackRegistry = CallbackRegistryReraise


@pytest.fixture(autouse=True)
def close_all():
    """Close all matplotlib plots, regardless of test status."""
    # This adds < 1 µS in local testing, and we have ~2500 tests, so ~2 ms max
    import matplotlib.pyplot as plt

    yield
    plt.close("all")


# We can't use monkeypatch because its scope (function-level) conflicts with
# the requests fixture (module-level), so we live with a module-scoped version
# that uses mock
@pytest.fixture(scope="module")
def options_3d():
    """Disable advanced 3d rendering."""
    with mock.patch.dict(
        os.environ,
        {
            "MNE_3D_OPTION_ANTIALIAS": "false",
            "MNE_3D_OPTION_DEPTH_PEELING": "false",
            "MNE_3D_OPTION_SMOOTH_SHADING": "false",
        },
    ):
        yield


@pytest.fixture
def requires_pyvista(options_3d):
    """Require pyvista."""
    if not mne.datasets.has_dataset("testing"):
        pytest.skip("Requires mne-testing-data")
    pyvista = pytest.importorskip("pyvista")
    pytest.importorskip("pyvistaqt")
    pyvista.close_all()
    try:
        from pyvista.plotting.plotter import _ALL_PLOTTERS
    except Exception:  # PV < 0.40
        from pyvista.plotting.plotting import _ALL_PLOTTERS
    assert len(_ALL_PLOTTERS) == 0
    mne.viz.set_3d_backend("pyvista")
    yield pyvista
    pyvista.close_all()
    assert len(_ALL_PLOTTERS) == 0



# Authors: Robert Luke <mail@robertluke.net>
#
# License: BSD (3-clause)

import datetime
import re

import h5py as h5py
import numpy as np
from mne.channels import make_standard_montage
from mne.io.pick import _picks_to_idx
from mne.transforms import _get_trans, apply_trans

# The currently-implemented spec can be found here:
# https://raw.githubusercontent.com/fNIRS/snirf/v1.1/snirf_specification.md
SPEC_FORMAT_VERSION = "1.1"


def write_raw_snirf(raw, fname, add_montage=False):
    """Write continuous wave data to disk in SNIRF format.

    Writes the data from a raw object to a SNIRF file using
    SNIRF specification version 1.1.
    The data can be either raw continuous wave amplitude,
    optical density or haemoglobin data.
    The data is written to the SNIRF file as a single data block
    and should pass the validation provided by the official
    snirf validator.

    The MNE SNIRF reader requires that the data file contains either
    two wavelengths or both hbo and hbr data. This function can export
    data with just one wavelength or just hbo or hbr data.
    However, until modifications are made to the SNIRF reader, the
    exported data will not be readable by the MNE SNIRF reader.

    Parameters
    ----------
    raw : instance of Raw
        Data to write to file. Can be either continuous wave
        (`fnirs_cw_amplitude`) or processed (`fnirs_od`, 'hbo', 'hbr').
    fname : str
        Path to the SNIRF data file.
    add_montage : bool
        Should the montage be added as landmarks to
        facilitate compatibility with AtlasViewer.
    """
    supported_types = ["fnirs_cw_amplitude", "fnirs_od", "hbo", "hbr"]
    picks = _picks_to_idx(raw.info, supported_types, exclude=[])
    assert len(picks) == len(raw.ch_names), (
        "Data must only be of type fnirs_cw_amplitude,  fnirs_od, hbo or hbr"
    )
    if ("fnirs_cw_amplitude" in raw) or ("fnirs_od" in raw):
        assert len(np.unique(raw.info.get_channel_types())) == 1, (
            "All channels must be of the same type"
        )
    elif ("hbo" in raw) or ("hbr" in raw):
        assert len(np.unique(raw.info.get_channel_types())) <= 2, (
            "Channels must be of type hbo and hbr"
        )

    with h5py.File(fname, "w") as f:
        nirs = f.create_group("/nirs")
        f.create_dataset("formatVersion", data=_str_encode(SPEC_FORMAT_VERSION))

        _add_metadata_tags(raw, nirs)
        _add_single_data_block(raw, nirs)
        _add_probe_info(raw, nirs, add_montage=add_montage)
        _add_stim_info(raw, nirs)


def _str_encode(str_val):
    """Encode a string for use in an h5py Dataset.

    Parameters
    ----------
    str_val : str
        The string to encode.
    """
    return str_val.encode("UTF-8")


def _add_metadata_tags(raw, nirs):
    """Create and add elements to the nirs metaDataTags group.

    Parameters
    ----------
    raw : instance of Raw
        Data to add to the snirf file.
    nirs : hpy5.Group
        The root hdf5 nirs group to which the metadata should beadded.
    """
    metadata_tags = nirs.create_group("metaDataTags")

    # Store measurement
    datestr = raw.info["meas_date"].strftime("%Y-%m-%d")
    timestr = raw.info["meas_date"].strftime("%H:%M:%SZ")
    metadata_tags.create_dataset("MeasurementDate", data=_str_encode(datestr))
    metadata_tags.create_dataset("MeasurementTime", data=_str_encode(timestr))

    # Store demographic info
    subject_id = "_".join(
        raw.info["subject_info"][key]
        for key in ["first_name", "middle_name", "last_name"]
        if key in raw.info["subject_info"]
    )
    subject_id = raw.info["subject_info"].get("his_id", subject_id)
    metadata_tags.create_dataset("SubjectID", data=_str_encode(subject_id))

    # Store the units of measurement
    metadata_tags.create_dataset("LengthUnit", data=_str_encode("m"))
    metadata_tags.create_dataset("TimeUnit", data=_str_encode("s"))
    metadata_tags.create_dataset("FrequencyUnit", data=_str_encode("Hz"))

    # Add non standard (but allowed) custom metadata tags
    if raw.info["subject_info"].get("birthday", None) is not None:
        birthday = raw.info["subject_info"]["birthday"]
        if isinstance(birthday, tuple):
            birthday = datetime.date(*birthday)
        birthstr = birthday.strftime("%Y-%m-%d")
        metadata_tags.create_dataset("DateOfBirth", data=[_str_encode(birthstr)])
    for key in ("first", "middle", "last"):
        name = raw.info["subject_info"].get(f"{key}_name", None)
        if name is not None:
            metadata_tags.create_dataset(f"{key}Name", data=[_str_encode(name)])
    if "sex" in raw.info["subject_info"]:
        sex = str(int(raw.info["subject_info"]["sex"]))
        metadata_tags.create_dataset("sex", data=[_str_encode(sex)])
    if raw.info["dig"] is not None:
        coord_frame_id = int(raw.info["dig"][0].get("coord_frame"))
        metadata_tags.create_dataset("MNE_coordFrame", data=[coord_frame_id])


def _add_single_data_block(raw, nirs):
    """Add the data from raw to the nirs data1 group.

    While SNIRF supports multiple datablocks, this writer only supports
    a single data block named data1.

    Parameters
    ----------
    raw : instance of Raw
        Data to add to the snirf file.
    nirs : hpy5.Group
        The root hdf5 nirs group to which the data should be added.
    """
    data_block = nirs.create_group("data1")
    data_block.create_dataset("dataTimeSeries", data=raw.get_data().T)
    data_block.create_dataset("time", data=raw.times)

    _add_measurement_lists(raw, data_block)


def _add_measurement_lists(raw, data_block):
    """Add the measurement list groups to the nirs data1 group.

    Parameters
    ----------
    raw : instance of Raw
        Data to add to the snirf file.
    data_block : hpy5.Group
        The hdf5 data1 group to which the measurement lists should be added.
    """
    sources = _get_unique_source_list(raw)
    detectors = _get_unique_detector_list(raw)
    wavelengths = _get_unique_wavelength_list(raw)

    raw_types = raw.info.get_channel_types()
    raw_wavelengths = [c["loc"][9] for c in raw.info["chs"]]

    for idx, ch_name in enumerate(raw.ch_names, start=1):
        ml_id = f"measurementList{idx}"
        ch_group = data_block.require_group(ml_id)

        source_idx = sources.index(_extract_source(ch_name)) + 1
        detector_idx = detectors.index(_extract_detector(ch_name)) + 1
        wavelength_idx = wavelengths.index(raw_wavelengths[idx - 1]) + 1
        ch_group.create_dataset("sourceIndex", data=source_idx, dtype="int32")
        ch_group.create_dataset("detectorIndex", data=detector_idx, dtype="int32")
        ch_group.create_dataset("wavelengthIndex", data=wavelength_idx, dtype="int32")

        # The data type coding is described at
        # https://github.com/fNIRS/snirf/blob/master/snirf_specification.md#appendix
        # The currently implemented data types are:
        # 1 = Continuous Wave
        # 99999 = Processed
        data_type = 1 if raw_types[idx - 1] == "fnirs_cw_amplitude" else 99999

        ch_group.create_dataset("dataType", data=data_type, dtype="int32")
        ch_group.create_dataset("dataTypeIndex", data=1, dtype="int32")
        if raw_types[idx - 1] == "fnirs_od":
            ch_group.create_dataset("dataTypeLabel", data="dOD")
        elif raw_types[idx - 1] == "fnirs_cw_amplitude":
            # The SNIRF specification does not specify a label for this type
            continue
        elif raw_types[idx - 1] == "hbo":
            ch_group.create_dataset("dataTypeLabel", data="HbO")
        elif raw_types[idx - 1] == "hbr":
            ch_group.create_dataset("dataTypeLabel", data="HbR")


def _add_probe_info(raw, nirs, add_montage):
    """Add details of the probe to the nirs group.

    Parameters
    ----------
    raw : instance of Raw
        Data to add to the snirf file.
    nirs : hpy5.Group
        The root hdf5 nirs group to which the probe info should be added.
    add_montage : bool
        Should the montage be added as landmarks to
        facilitate compatibility with AtlasViewer.
    """
    sources = _get_unique_source_list(raw)
    detectors = _get_unique_detector_list(raw)
    wavelengths = _get_unique_wavelength_list(raw)

    probe = nirs.create_group("probe")

    # Store source/detector/wavelength info
    encoded_source_labels = [_str_encode(f"S{src}") for src in sources]
    encoded_detector_labels = [_str_encode(f"D{det}") for det in detectors]
    probe.create_dataset("sourceLabels", data=encoded_source_labels)
    probe.create_dataset("detectorLabels", data=encoded_detector_labels)
    probe.create_dataset("wavelengths", data=wavelengths)

    # Create 3d locs and store
    ch_sources = [_extract_source(ch) for ch in raw.ch_names]
    ch_detectors = [_extract_detector(ch) for ch in raw.ch_names]
    srclocs = np.empty((len(sources), 3))
    detlocs = np.empty((len(detectors), 3))
    for i, src in enumerate(sources):
        idx = ch_sources.index(src)
        srclocs[i, :] = raw.info["chs"][idx]["loc"][3:6]
    for i, det in enumerate(detectors):
        idx = ch_detectors.index(det)
        detlocs[i, :] = raw.info["chs"][idx]["loc"][6:9]
    probe.create_dataset("sourcePos3D", data=srclocs)
    probe.create_dataset("detectorPos3D", data=detlocs)

    # Store probe landmarks
    if raw.info["dig"] is not None:
        _store_probe_landmarks(raw, probe, add_montage)


def _store_probe_landmarks(raw, probe, add_montage):
    """Add the probe landmarks to the probe group.

    The SNIRF specification provides flexibility around
    what is stored in this field. Some software expect specific
    landmarks which are not part of the SNIRF specification.
    We endeavour to provide as much information as possible,
    without causing conflicts with the specification,
    in order to achieve compliance with other software packages.

    Atlas Viewer expects a list of 10-20 locations to be provided
    in the landmark fields.

    We also store the values in `raw.info['dig']` as these
    may represent positions that have been digitised by the
    researcher and should be retained. These are prepended
    by the string "HP" to represent a head point.

    Parameters
    ----------
    raw : instance of Raw
        Data to add to the snirf file.
    probe : hpy5.Group
        The hdf5 probe group to which the landmark info should be added.
    add_montage : bool
        Should the montage be added as landmarks to
        facilitate compatibility with AtlasViewer.
    """
    diglocs = np.empty((len(raw.info["dig"]), 3))
    digname = list()
    for idx, dig in enumerate(raw.info["dig"]):
        ident = re.match(r"\d+ \(FIFFV_POINT_(\w+)\)", str(dig.get("ident")))
        if ident is not None:
            digname.append(ident[1])
        else:
            digname.append(f"HP_{str(dig.get('ident'))}")
        diglocs[idx, :] = dig.get("r")

    if add_montage:
        # First, get the template positions in mni fsaverage space
        montage = make_standard_montage("standard_1020", head_size=0.09700884729534559)
        ch_names = montage.ch_names
        montage_locs = np.array(list(montage._get_ch_pos().values()))

        # montage_locs = np.array([d['r'] for d in montage.dig])
        head_mri_t, _ = _get_trans("fsaverage", "mri", "head")
        locations_head = apply_trans(head_mri_t, montage_locs)

        digname.extend(ch_names)
        diglocs = np.concatenate((diglocs, locations_head))

    digname = [_str_encode(d) for d in digname]
    probe.create_dataset("landmarkPos3D", data=diglocs)
    probe.create_dataset("landmarkLabels", data=digname)


def _add_stim_info(raw, nirs):
    """Add details of the stimuli to the nirs group.

    Parameters
    ----------
    raw : instance of Raw
        Data to add to the snirf file.
    nirs : hpy5.Group
        The root hdf5 nirs group to which the stimuli info should be added.
    """
    # Convert MNE annotations to SNIRF stims
    descriptions = np.unique(raw.annotations.description)
    for idx, desc in enumerate(descriptions, start=1):
        stim_group = nirs.create_group(f"stim{idx}")
        trgs = np.where(raw.annotations.description == desc)[0]
        stims = np.zeros((len(trgs), 3))
        for idx_t, trg in enumerate(trgs):
            stims[idx_t, :] = [
                raw.annotations.onset[trg],
                raw.annotations.duration[trg],
                idx,
            ]
        stim_group.create_dataset("data", data=stims)
        stim_group.create_dataset("name", data=_str_encode(desc))


def _get_unique_source_list(raw):
    """Return the sorted list of distinct source ids.

    Parameters
    ----------
    raw : instance of Raw
        Data object containing the list of channels.
    """
    ch_sources = [_extract_source(ch_name) for ch_name in raw.ch_names]
    return list(sorted(set(ch_sources)))


def _get_unique_detector_list(raw):
    """Return the sorted list of distinct detector ids.

    Parameters
    ----------
    raw : instance of Raw
        Data object containing the list of channels.
    """
    ch_detectors = [_extract_detector(ch_name) for ch_name in raw.ch_names]
    return list(sorted(set(ch_detectors)))


def _get_unique_wavelength_list(raw):
    """Return the sorted list of distinct wavelengths.

    Parameters
    ----------
    raw : instance of Raw
        Data object containing the list of channels.
    """
    ch_wavelengths = [c["loc"][9] for c in raw.info["chs"]]
    return list(sorted(set(ch_wavelengths)))


def _match_channel_pattern(channel_name):
    """Return a regex match against the expected channel name format.

    The returned match object contains three named groups: source, detector,
    and wavelength/type. If no match is found, a ValueError is raised.

    For example, if the name is "S1_D2 hbo" the source is 1,
    the detector is 2, and the type is hbo.
    If the name is "S1_D2 850" the source is 1,
    the detector is 2, and the wavelength is 850.

    Parameters
    ----------
    channel_name : str
        The name of the channel.
    """
    rgx = r"^S(?P<source>\d+)_D(?P<detector>\d+) (?P<wavelength_type>[\w]+)$"
    match = re.fullmatch(rgx, channel_name)
    if match is None:
        msg = f"channel name does not match expected pattern: {channel_name}"
        raise ValueError(msg)
    return match


def _extract_source(channel_name):
    """Extract and return the source id from the channel name.

    The id is returned as an integer value.

    Parameters
    ----------
    channel_name : str
        The name of the channel.
    """
    return int(_match_channel_pattern(channel_name).group("source"))


def _extract_detector(channel_name):
    """Extract and return the detector id from the channel name.

    The id is returned as an integer value.

    Parameters
    ----------
    channel_name : str
        The name of the channel.
    """
    return int(_match_channel_pattern(channel_name).group("detector"))



# Authors: Robert Luke <mail@robertluke.net>
#
# License: BSD (3-clause)

import mne
import numpy as np


def make_first_level_design_matrix(
    raw,
    stim_dur=1.0,
    hrf_model="glover",
    drift_model="cosine",
    high_pass=0.01,
    drift_order=1,
    fir_delays=(0,),
    add_regs=None,
    add_reg_names=None,
    min_onset=-24,
    oversampling=50,
):
    """
    Generate a design matrix based on annotations and model HRF.

    This is a wrapper function for the nilearn :footcite:`abraham2014machine`
    function ``make_first_level_design_matrix``. For detailed description
    of the arguments see the nilearn documentation at http://nilearn.github.io

    Parameters
    ----------
    raw : instance of Raw
        Haemoglobin data.

    stim_dur : Number
        The length of your stimulus.

    hrf_model : {'glover', 'spm', 'spm + derivative', \
        'spm + derivative + dispersion',\
        'glover + derivative', 'glover + derivative + dispersion',\
        'fir', None}, optional
        Specifies the hemodynamic response function. Default='glover'.

    drift_model : {'cosine', 'polynomial', None}, optional
        Specifies the desired drift model. Default='cosine'.

    high_pass : float, optional
        High-pass frequency in case of a cosine model (in Hz).
        Default=0.01.

    drift_order : int, optional
        Order of the drift model (in case it is polynomial).
        Default=1.

    fir_delays : array of shape(n_onsets) or list, optional
        In case of FIR design, yields the array of delays used in the FIR
        model (in scans). Default=[0].

    add_regs : array of shape(n_frames, n_add_reg) or pandas DataFrame
        Additional user-supplied regressors, e.g. data driven noise regressors
        or seed based regressors.

    add_reg_names : list of (n_add_reg,) str, optional
        If None, while add_regs was provided, these will be termed
        'reg_%i', i = 0..n_add_reg - 1
        If add_regs is a DataFrame, the corresponding column names are used
        and add_reg_names is ignored.

    min_onset : float, optional
        Minimal onset relative to frame_times[0] (in seconds)
        events that start before frame_times[0] + min_onset are not considered.
        Default=-24.

    oversampling : int, optional
        Oversampling factor used in temporal convolutions. Default=50.

    Returns
    -------
    design_matrix : DataFrame instance,
        Holding the computed design matrix, the index being the frames_times
        and each column a regressor.

    References
    ----------
    .. footbibliography::
    """
    from nilearn.glm.first_level import make_first_level_design_matrix
    from pandas import DataFrame

    frame_times = raw.times

    # Create events for nilearn
    conditions = raw.annotations.description
    onsets = raw.annotations.onset - raw.first_time
    duration = stim_dur * np.ones(len(conditions))
    events = DataFrame(
        {"trial_type": conditions, "onset": onsets, "duration": duration}
    )

    dm = make_first_level_design_matrix(
        frame_times,
        events,
        drift_model=drift_model,
        drift_order=drift_order,
        hrf_model=hrf_model,
        min_onset=min_onset,
        high_pass=high_pass,
        add_regs=add_regs,
        oversampling=oversampling,
        add_reg_names=add_reg_names,
        fir_delays=fir_delays,
    )

    return dm


def create_boxcar(raw, event_id=None, stim_dur=1):
    """
    Generate boxcar representation of the experimental paradigm.

    Parameters
    ----------
    raw : instance of Raw
        Haemoglobin data.
    event_id : as specified in MNE
        Information about events.
    stim_dur : Number
        The length of your stimulus.

    Returns
    -------
    s : array
        Returns an array for each annotation label.
    """
    bc = np.ones(int(round(raw.info["sfreq"] * stim_dur)))
    events, ids = mne.events_from_annotations(raw, event_id=event_id)
    s = np.zeros((len(raw.times), len(ids)))
    for idx, _ in enumerate(ids):
        id_idx = [e[2] == idx + 1 for e in events]
        id_evt = events[id_idx]
        event_samples = [e[0] for e in id_evt]
        s[event_samples, idx] = 1.0
        s[:, idx] = np.convolve(s[:, idx], bc)[: len(raw.times)]
    return s


def longest_inter_annotation_interval(raw):
    """
    Compute longest ISI per annotation.

    Specifically, longest period between two trials of
    the same condition.

    Parameters
    ----------
    raw : instance of Raw
        Haemoglobin data.

    Returns
    -------
    longest : list
        Longest ISI per annotation.
    annotation_name : list
        Annotation name corresponding to reported interval.
    """
    annotation_name = np.unique(raw.annotations.description)
    longest = []
    for desc in annotation_name:
        mask = raw.annotations.description == desc
        longest.append(np.max(np.diff(raw.annotations.onset[mask])))
    return longest, annotation_name


def drift_high_pass(raw):
    """
    Compute cosine drift regressor high pass cut off.

    Value computed according to Nilearn :footcite:`abraham2014machine`
    `suggestion <http://nilearn.github.io/auto_examples/04_glm_first
    _level/plot_first_level_details.html#changing-the-drift-model>`__.

    Parameters
    ----------
    raw : instance of Raw
        Haemoglobin data.

    Returns
    -------
    cutoff : number
        Suggested high pass cut off.

    References
    ----------
    .. footbibliography::
    """
    longest, annotation_name = longest_inter_annotation_interval(raw)
    max_isi = np.max(longest)
    return 1 / (2 * max_isi)



# Authors: Eric Larson <larson.eric.d@gmail.com>
# License: BSD (3-clause)
#
# Compat shims for different dependency versions.

from pathlib import Path

import mne
from mne.utils import check_version

# Deal with fetch_data str->Path deprecation
# This can go away once we only support 1.1, along with the
# conftest.py ignore of "data_path functions now return"
try:
    from mne.datasets.utils import _mne_path  # noqa
except Exception:
    if check_version(mne.__version__, "1.0"):
        _mne_path = Path
    else:
        _mne_path = str  # old MNE



# Authors: Robert Luke <mail@robertluke.net>
#
# License: BSD (3-clause)


import os

import mne
from numpy.testing import assert_array_equal

from mne_nirs.channels import (
    drop_detectors,
    drop_sources,
    list_detectors,
    list_sources,
    pick_detectors,
    pick_sources,
)


def _get_raw():
    fnirs_data_folder = mne.datasets.fnirs_motor.data_path()
    fnirs_raw_dir = os.path.join(fnirs_data_folder, "Participant-1")
    raw = mne.io.read_raw_nirx(fnirs_raw_dir).load_data()
    return raw


def test_list_sources():
    raw = _get_raw()

    sources = list_sources(raw)
    assert len(sources) == 8
    assert_array_equal(sources, range(1, 9))

    sources = list_sources(raw.copy().pick("S3_D2 760"))
    assert sources == [3]

    sources = list_sources(raw.copy().pick("S7_D15 850"))
    assert sources == [7]

    sources = list_sources(raw.copy().pick(["S7_D15 850", "S7_D15 760"]))
    assert sources == [7]

    sources = list_sources(raw.copy().pick(["S7_D15 850", "S3_D2 850", "S7_D15 760"]))
    assert_array_equal(sources, [3, 7])


def test_list_detectors():
    raw = _get_raw()

    detectors = list_detectors(raw)
    assert len(detectors) == 16
    assert_array_equal(detectors, range(1, 17))

    detectors = list_detectors(raw.copy().pick("S3_D2 760"))
    assert detectors == [2]

    detectors = list_detectors(raw.copy().pick("S7_D15 850"))
    assert detectors == [15]

    detectors = list_detectors(raw.copy().pick(["S7_D15 850", "S7_D15 760"]))
    assert detectors == [15]

    detectors = list_detectors(
        raw.copy().pick(["S7_D15 850", "S3_D2 850", "S7_D15 760"])
    )
    assert_array_equal(detectors, [2, 15])


def test_drop_sources():
    raw = _get_raw()

    sources = list_sources(raw)
    assert_array_equal(sources, range(1, 9))

    raw_dropped = drop_sources(raw.copy(), 3)
    assert 3 not in list_sources(raw_dropped)

    raw_dropped = drop_sources(raw.copy(), [7, 2])
    assert 7 not in list_sources(raw_dropped)
    assert 2 not in list_sources(raw_dropped)


def test_drop_detectors():
    raw = _get_raw()

    detectors = list_detectors(raw)
    assert_array_equal(detectors, range(1, 17))

    raw_dropped = drop_detectors(raw.copy(), 3)
    assert 3 not in list_detectors(raw_dropped)

    raw_dropped = drop_detectors(raw.copy(), [7, 2])
    assert 7 not in list_detectors(raw_dropped)
    assert 2 not in list_detectors(raw_dropped)

    raw_dropped = drop_detectors(raw.copy(), [1, 2])
    assert 12 in list_detectors(raw_dropped)
    assert 2 not in list_detectors(raw_dropped)
    assert 1 not in list_detectors(raw_dropped)


def test_pick_sources():
    raw = _get_raw()

    sources = list_sources(raw)
    assert_array_equal(sources, range(1, 9))

    raw_picked = pick_sources(raw.copy(), 3)
    assert 3 in list_sources(raw_picked)
    assert 2 not in list_sources(raw_picked)
    assert 1 not in list_sources(raw_picked)

    raw_picked = pick_sources(raw.copy(), [7, 2])
    assert 7 in list_sources(raw_picked)
    assert 2 in list_sources(raw_picked)
    assert 1 not in list_sources(raw_picked)
    assert 11 not in list_sources(raw_picked)
    assert 9 not in list_sources(raw_picked)


def test_pick_detectors():
    raw = _get_raw()

    sources = list_detectors(raw)
    assert_array_equal(sources, range(1, 17))

    raw_picked = pick_detectors(raw.copy(), 3)
    assert 3 in list_detectors(raw_picked)
    assert 2 not in list_detectors(raw_picked)
    assert 1 not in list_detectors(raw_picked)

    raw_picked = pick_detectors(raw.copy(), [7, 2])
    assert 7 in list_detectors(raw_picked)
    assert 2 in list_detectors(raw_picked)
    assert 1 not in list_detectors(raw_picked)
    assert 11 not in list_detectors(raw_picked)
    assert 9 not in list_detectors(raw_picked)



# Authors: Robert Luke <mail@robertluke.net>
#
# License: BSD (3-clause)

from ._experimental_design import (
    create_boxcar,
    make_first_level_design_matrix,
    longest_inter_annotation_interval,
    drift_high_pass,
)



# Authors: Robert Luke <mail@robertluke.net>
#
# License: BSD (3-clause)


import numpy as np
from mne.utils import warn


def picks_pair_to_idx(raw, sd_pairs, on_missing="error"):
    """
    Return a list of picks for specified source detector pairs.

    If multiple channel have the same source detector pair,
    for example if there are multiple wavelengths or chromaphore, then
    all channels matching the requested sources and detectors will be
    returned.

    Parameters
    ----------
    raw : instance of Raw
        The haemoglobin data.
    sd_pairs : list of lists
        List of source detector pairs. For example, to request the picks for
        channels comprising of source 1 detector 4 and source 13
        detector 4 you would specify [[1, 4], [13, 4]].
    on_missing : str
        What to do if one or several requested source detector are not found
        in the recording.
        Valid keys are 'error' | 'warning' | 'ignore'
        Default is 'error'. If on_missing is 'warning' it will proceed but
        warn, if 'ignore' it will proceed silently. Note.
        If none of the event ids are found in the data, an error will be
        automatically generated irrespective of this parameter.

    Returns
    -------
    picks : list of integers
        List of picks corresponding to requested source detector pairs.
    """
    ch_names = raw.ch_names
    picks = list()

    for pair in sd_pairs:
        pair_name = "S" + str(pair[0]) + "_D" + str(pair[1]) + " "
        pair_picks = np.where([pair_name in ch for ch in ch_names])[0]
        if len(pair_picks) == 0:
            msg = "No matching channels found for source %s detector %s" % (
                pair[0],
                pair[1],
            )
            if on_missing == "error":
                print(pair_picks)
                raise ValueError(msg)
            elif on_missing == "warning":
                warn(msg)
            else:
                # on_missing == 'ignore':
                continue
        [picks.append(pick) for pick in pair_picks]

    return picks



# Authors: Robert Luke <mail@robertluke.net>
#
# License: BSD (3-clause)

from ._roi import glm_region_of_interest
from ._statsmodels import statsmodels_to_results
from ._glm_level_first import (
    RegressionResults,
    ContrastResults,
    run_glm,
    run_GLM,
    read_glm,
)



# Authors: Robert Luke <mail@robertluke.net>
#
# License: BSD (3-clause)

import os

import mne

from mne_nirs.preprocessing import peak_power, scalp_coupling_index_windowed


def test_peak_power():
    fnirs_data_folder = mne.datasets.fnirs_motor.data_path()
    fnirs_raw_dir = os.path.join(fnirs_data_folder, "Participant-1")
    raw = mne.io.read_raw_nirx(fnirs_raw_dir, verbose=True).load_data()
    raw = mne.preprocessing.nirs.optical_density(raw)

    raw, scores, times = peak_power(raw)
    assert len(scores) == len(raw.ch_names)


def test_sci_windowed():
    fnirs_data_folder = mne.datasets.fnirs_motor.data_path()
    fnirs_raw_dir = os.path.join(fnirs_data_folder, "Participant-1")
    raw = mne.io.read_raw_nirx(fnirs_raw_dir, verbose=True).load_data()
    raw = mne.preprocessing.nirs.optical_density(raw)

    raw, scores, times = scalp_coupling_index_windowed(raw)
    assert len(scores) == len(raw.ch_names)



import re

import nilearn
import numpy as np
import pandas as pd
from mne.utils import warn
from scipy import stats


def glm_to_tidy(info, statistic, design_matrix, wide=True, order=None):
    """
    Export GLM regression or contrast results in tidy format.

    Creates a long pandas data frame from regression results or contrast
    as computed by run_glm or compute_contrast.

    Parameters
    ----------
    info : MNE.Info
        Instance of MNE info.
    statistic : nilearn data,
        Either dict of nilearn.stats.regression.RegressionResults as returned
        by run_glm, or nilearn.stats.contrasts.Contrast as returned by
        compute_contrast.
    design_matrix : DataFrame
        As specified in Nilearn.
    wide : Bool
        Should the returned dataframe be in wide format. If False, then the
        returned data will be in long format.
    order : list
        Order that the channels should be returned with.

    Returns
    -------
    df : Tidy data frame,
        Data from statistic object in tidy data form.
    """
    if isinstance(statistic, dict) and isinstance(
        statistic[list(statistic.keys())[0]], nilearn.glm.regression.RegressionResults
    ):
        df = _tidy_RegressionResults(info, statistic, design_matrix)

    elif isinstance(statistic, nilearn.glm.contrasts.Contrast):
        df = _tidy_Contrast(info, statistic, design_matrix)

    else:
        raise TypeError(
            "Unknown statistic type. Expected dict of RegressionResults "
            f"or Contrast type. Received {type(statistic)}"
        )

    if wide:
        df = _tidy_long_to_wide(df, expand_output=True)

    if order is not None:
        df["old_index"] = df.index
        df = df.set_index("ch_name")
        df = df.loc[order, :]
        df["ch_name"] = df.index
        df.index = df["old_index"]
        df.drop(columns="old_index", inplace=True)
        df.rename_axis(None, inplace=True)

    return df


def _tidy_Contrast(data, glm_est, design_matrix):
    df = pd.DataFrame()
    if hasattr(glm_est, "stat_type"):
        contrast_type = glm_est.stat_type
    else:  # nilearn < 0.10
        contrast_type = glm_est.contrast_type
    effect = glm_est.effect
    if effect.ndim == 2:  # old nilearn
        assert effect.shape[0] == 1
        effect = effect[0]
    for idx, ch in enumerate(data.ch_names):
        df = pd.concat(
            [
                df,
                pd.DataFrame(
                    {
                        "ch_name": ch,
                        "ContrastType": contrast_type,
                        "variable": "effect",
                        "value": effect[idx],
                    },
                    index=[0],
                ),
                pd.DataFrame(
                    {
                        "ch_name": ch,
                        "ContrastType": contrast_type,
                        "variable": "p_value",
                        "value": glm_est.p_value()[idx],
                    },
                    index=[1],
                ),
                pd.DataFrame(
                    {
                        "ch_name": ch,
                        "ContrastType": contrast_type,
                        "variable": "stat",
                        "value": glm_est.stat()[idx],
                    },
                    index=[2],
                ),
                pd.DataFrame(
                    {
                        "ch_name": ch,
                        "ContrastType": contrast_type,
                        "variable": "z_score",
                        "value": glm_est.z_score()[idx],
                    },
                    index=[3],
                ),
            ],
            ignore_index=True,
        )
    return df


def _tidy_RegressionResults(data, glm_est, design_matrix):
    if not (data.ch_names == list(glm_est.keys())):
        warn("MNE data structure does not match regression results")

    theta_estimates = np.zeros((len(glm_est), len(design_matrix.columns)))
    t_estimates = np.zeros((len(glm_est), len(design_matrix.columns)))
    df_estimates = np.zeros((len(glm_est), len(design_matrix.columns)))
    p_estimates = np.zeros((len(glm_est), len(design_matrix.columns)))
    mse_estimates = np.zeros((len(glm_est), len(design_matrix.columns)))
    se_estimates = np.zeros((len(glm_est), len(design_matrix.columns)))

    for idx, name in enumerate(glm_est.keys()):
        theta_estimates[idx, :] = glm_est[name].theta.T
        df_estimates[idx, :] = glm_est[name].df_model
        mse_estimates[idx, :] = glm_est[name].MSE[0]
        for cond_idx, cond in enumerate(design_matrix.columns):
            t_estimates[idx, cond_idx] = glm_est[name].t(column=cond_idx).item()
            p_estimates[idx, cond_idx] = 2 * stats.t.cdf(
                -1.0 * np.abs(t_estimates[idx, cond_idx]),
                df=df_estimates[idx, cond_idx],
            )
            se_estimates[idx, cond_idx] = np.sqrt(np.diag(glm_est[name].vcov()))[
                cond_idx
            ]

    list_vals = [0] * (len(data.ch_names) * len(design_matrix.columns) * 6)
    idx = 0
    for ch_idx, ch in enumerate(data.ch_names):
        for cond_idx, cond in enumerate(design_matrix.columns):
            list_vals[0 + idx] = {
                "ch_name": ch,
                "Condition": cond,
                "variable": "theta",
                "value": theta_estimates[ch_idx][cond_idx],
            }
            list_vals[1 + idx] = {
                "ch_name": ch,
                "Condition": cond,
                "variable": "t",
                "value": t_estimates[ch_idx][cond_idx],
            }
            list_vals[2 + idx] = {
                "ch_name": ch,
                "Condition": cond,
                "variable": "df",
                "value": df_estimates[ch_idx][cond_idx],
            }
            list_vals[3 + idx] = {
                "ch_name": ch,
                "Condition": cond,
                "variable": "p_value",
                "value": p_estimates[ch_idx][cond_idx],
            }
            list_vals[4 + idx] = {
                "ch_name": ch,
                "Condition": cond,
                "variable": "mse",
                "value": mse_estimates[ch_idx][cond_idx],
            }
            list_vals[5 + idx] = {
                "ch_name": ch,
                "Condition": cond,
                "variable": "se",
                "value": se_estimates[ch_idx][cond_idx],
            }
            idx += 6

    dict_vals, i = {}, 0
    for entry in list_vals:
        dict_vals[i] = {
            "ch_name": entry["ch_name"],
            "Condition": entry["Condition"],
            "variable": entry["variable"],
            "value": entry["value"],
        }
        i = i + 1
    df = pd.DataFrame.from_dict(dict_vals, "index")

    return df


def _tidy_long_to_wide(d, expand_output=True):
    indices = ["ch_name"]
    if "Condition" in d.columns:
        # Regression results have a column condition
        indices.append("Condition")
    if "ContrastType" in d.columns:
        # Regression results have a column condition
        indices.append("ContrastType")

    d = d.set_index(indices)
    d = d.pivot_table(columns="variable", values="value", index=indices)
    d.reset_index(inplace=True)

    if expand_output:
        try:
            d["Source"] = [
                int(re.search(r"S(\d+)_D(\d+) (\w+)", ch).group(1))
                for ch in d["ch_name"]
            ]
            d["Detector"] = [
                int(re.search(r"S(\d+)_D(\d+) (\w+)", ch).group(2))
                for ch in d["ch_name"]
            ]
            d["Chroma"] = [
                re.search(r"S(\d+)_D(\d+) (\w+)", ch).group(3) for ch in d["ch_name"]
            ]
        except AttributeError:
            warn("Non standard source detector names used")
        d["Significant"] = d["p_value"] < 0.05

    return d



# Authors: Robert Luke <mail@robertluke.net>
#
# License: BSD (3-clause)

from . import snirf
from . import fold
from .snirf import write_raw_snirf, read_snirf_aux_data
from .fold import fold_channel_specificity, fold_landmark_specificity



from importlib.metadata import version

try:
    __version__ = version("mne-nirs")
except Exception:
    __version__ = "0.0.0"
del version


from . import channels
from . import datasets
from . import experimental_design
from . import io
from . import preprocessing
from . import signal_enhancement
from . import simulation
from . import statistics
from . import utils
from . import visualisation
from . import visualisation as viz  # for MNE-Python users

__all__ = ["__version__"]



# Authors: Robert Luke <mail@robertluke.net>
#
# License: BSD (3-clause)

import os

import mne
import numpy as np
import pytest

from mne_nirs.preprocessing import quantify_mayer_fooof

pytest.importorskip("fooof")


# in fooof->scipy.optimize.curve_fit
@pytest.mark.filterwarnings("ignore:divide by zero encountered in divide.*:")
@pytest.mark.filterwarnings("ignore:invalid value encountered in.*:")
def test_mayer():
    fnirs_data_folder = mne.datasets.fnirs_motor.data_path()
    fnirs_raw_dir = os.path.join(fnirs_data_folder, "Participant-1")
    raw_intensity = mne.io.read_raw_nirx(fnirs_raw_dir, verbose=True).load_data()

    raw_intensity = raw_intensity.pick(picks=range(8)).crop(tmax=600)

    with pytest.raises(RuntimeError, match="run on haemoglobin"):
        _ = quantify_mayer_fooof(raw_intensity)
    raw_od = mne.preprocessing.nirs.optical_density(raw_intensity)
    with pytest.raises(RuntimeError, match="run on haemoglobin"):
        _ = quantify_mayer_fooof(raw_od)

    raw_haemo = mne.preprocessing.nirs.beer_lambert_law(raw_od, ppf=0.1)
    df_mayer = quantify_mayer_fooof(raw_haemo)

    print(df_mayer)

    assert df_mayer.shape[0] == 2
    assert df_mayer.shape[1] == 4
    assert (
        np.abs(df_mayer.query('Chromaphore == "hbo"')["Centre Frequency"][0] - 0.1)
        < 0.05
    )



# Authors: Robert Luke <mail@robertluke.net>
#
# License: BSD (3-clause)


import os

import mne
import numpy as np
import pytest

import mne_nirs
from mne_nirs.experimental_design import make_first_level_design_matrix
from mne_nirs.statistics import run_glm
from mne_nirs.statistics._glm_level_first import _compute_contrast
from mne_nirs.utils._io import _tidy_long_to_wide, glm_to_tidy


@pytest.mark.filterwarnings("ignore:.*more comprehensive.*:")
def test_io():
    num_chans = 6
    fnirs_data_folder = mne.datasets.fnirs_motor.data_path()
    fnirs_raw_dir = os.path.join(fnirs_data_folder, "Participant-1")
    raw_intensity = mne.io.read_raw_nirx(fnirs_raw_dir).load_data()
    raw_intensity.resample(0.2)
    raw_intensity.annotations.description[:] = [
        "e" + d.replace(".", "p") for d in raw_intensity.annotations.description
    ]
    raw_od = mne.preprocessing.nirs.optical_density(raw_intensity)
    raw_haemo = mne.preprocessing.nirs.beer_lambert_law(raw_od, ppf=0.1)
    raw_haemo = mne_nirs.channels.get_long_channels(raw_haemo)
    raw_haemo.pick(picks=range(num_chans))
    design_matrix = make_first_level_design_matrix(
        raw_intensity,
        hrf_model="spm",
        stim_dur=5.0,
        drift_order=3,
        drift_model="polynomial",
    )
    glm_est = run_glm(raw_haemo, design_matrix)
    df = glm_to_tidy(raw_haemo, glm_est.data, design_matrix)
    assert df.shape == (48, 12)
    assert set(df.columns) == {
        "ch_name",
        "Condition",
        "df",
        "mse",
        "p_value",
        "t",
        "theta",
        "Source",
        "Detector",
        "Chroma",
        "Significant",
        "se",
    }
    num_conds = 8  # triggers (1, 2, 3, 15) + 3 drifts + constant
    assert df.shape[0] == num_chans * num_conds
    assert len(df["se"]) == 48
    assert sum(df["se"]) > 0  # Check isn't nan
    assert len(df["df"]) == 48
    assert sum(df["df"]) > 0  # Check isn't nan
    assert len(df["p_value"]) == 48
    assert sum(df["p_value"]) > 0  # Check isn't nan
    assert len(df["theta"]) == 48
    assert sum(df["theta"]) > 0  # Check isn't nan
    assert len(df["t"]) == 48
    assert sum(df["t"]) > -99999  # Check isn't nan

    contrast_matrix = np.eye(design_matrix.shape[1])
    basic_conts = dict(
        [(column, contrast_matrix[i]) for i, column in enumerate(design_matrix.columns)]
    )
    contrast_LvR = basic_conts["e2p0"] - basic_conts["e3p0"]

    contrast = _compute_contrast(glm_est.data, contrast_LvR)
    df = glm_to_tidy(raw_haemo, contrast, design_matrix)
    assert df.shape == (6, 10)
    assert set(df.columns) == {
        "ch_name",
        "ContrastType",
        "z_score",
        "stat",
        "p_value",
        "effect",
        "Source",
        "Detector",
        "Chroma",
        "Significant",
    }

    contrast = _compute_contrast(glm_est.data, contrast_LvR, contrast_type="F")
    df = glm_to_tidy(raw_haemo, contrast, design_matrix, wide=False)
    df = _tidy_long_to_wide(df)
    assert df.shape == (6, 10)
    assert set(df.columns) == {
        "ch_name",
        "ContrastType",
        "z_score",
        "stat",
        "p_value",
        "effect",
        "Source",
        "Detector",
        "Chroma",
        "Significant",
    }

    with pytest.raises(TypeError, match="Unknown statistic type"):
        glm_to_tidy(raw_haemo, [1, 2, 3], design_matrix, wide=False)



# Authors: Robert Luke <code@robertluke.net>
# License: BSD (3-clause)
#
# This downloads SNIRF data that includes auxiliary channels.

import os
from functools import partial

import pooch
from mne.datasets import fetch_dataset
from mne.datasets.utils import has_dataset
from mne.utils import verbose

from ...fixes import _mne_path

has_fnirs_snirf_aux_data = partial(has_dataset, name="snirf_with_aux")


@verbose
def data_path(
    path=None, force_update=False, update_path=True, download=True, verbose=None
):  # noqa: D103
    """
    SNIRF file with auxiliary channels.

    Get path to local copy of data.

    Parameters
    ----------
    path : None | str
        Location of where to look for the dataset.
        If None, the environment variable or config parameter is used.
        If it doesn’t exist, the “~/mne_data” directory is used.
        If the dataset is not found under the given path,
        the data will be automatically downloaded to the specified folder.
    force_update : bool
        Force update of the dataset even if a local copy exists.
    update_path : bool | None
        If True, set the MNE_DATASETS_SNIRFAUX_PATH in
        mne-python config to the given path. If None, the user is prompted.
    download : bool
        If False and the dataset has not been downloaded yet,
        it will not be downloaded and the path will be returned
        as ‘’ (empty string). This is mostly used for debugging purposes
        and can be safely ignored by most users.
    %(verbose)s

    Returns
    -------
    path : str
        Path to dataset directory.
    """
    dataset_params = dict(
        archive_name="2022-08-05_002.snirf.zip",
        hash="md5:c012485b80d762a65b05d812761c0f4c",
        url="https://osf.io/download/682ddabd4d28c665accacc7c?version=2",
        folder_name="fNIRS-SNIRF-aux",
        dataset_name="snirf_with_aux",
        config_key="MNE_DATASETS_SNIRFAUX_PATH",
    )

    dpath = fetch_dataset(
        dataset_params,
        path=path,
        force_update=force_update,
        update_path=update_path,
        download=download,
        processor=pooch.Unzip(extract_dir="./fNIRS-SNIRF-aux"),
    )
    dpath = str(dpath)

    return _mne_path(os.path.join(dpath, "2022-08-05_002.snirf"))



# Authors: Robert Luke <mail@robertluke.net>
# License: BSD (3-clause)

import os.path as op

import mne_nirs


def test_dataset_av_speech():
    datapath = mne_nirs.datasets.audio_or_visual_speech.data_path()
    assert op.isdir(datapath)
    assert op.isdir(op.join(datapath, "sub-01"))

    # First pass downloaded, check that second pass of access works
    datapath = mne_nirs.datasets.audio_or_visual_speech.data_path()
    assert op.isdir(datapath)
    assert op.isdir(op.join(datapath, "sub-01"))



# Authors: Robert Luke <mail@robertluke.net>
#
# License: BSD (3-clause)

import numpy as np
from mne import pick_types
from mne.io import BaseRaw
from mne.preprocessing.nirs import source_detector_distances
from mne.utils import _validate_type
from scipy import linalg


def short_channel_regression(raw, max_dist=0.01):
    """
    Systemic correction regression based on nearest short channel.

    Method as described by NIRx and based on
    :footcite:`fabbri2004optical`, :footcite:`saager2005direct`,
    and :footcite:`scholkmann2014measuring`.

    Parameters
    ----------
    raw : instance of Raw
        Raw instance containing optical density data.
    max_dist : number
        Channels less than this distance are considered short (m).

    Returns
    -------
    raw : instance of Raw
        The modified raw instance.

    References
    ----------
    .. footbibliography::
    """
    raw = raw.copy().load_data()
    _validate_type(raw, BaseRaw, "raw")

    picks_od = pick_types(raw.info, fnirs="fnirs_od")

    if len(picks_od) == 0:
        raise RuntimeError("Data must be optical density.")

    distances = source_detector_distances(raw.info)

    picks_short = picks_od[distances[picks_od] < max_dist]
    picks_long = picks_od[distances[picks_od] > max_dist]

    if len(picks_short) == 0:
        raise RuntimeError("No short channels present.")
    if len(picks_long) == 0:
        raise RuntimeError("No long channels present.")

    for pick in picks_long:
        short_idx = _find_nearest_short(raw, pick, picks_short)

        A_l = raw.get_data(pick).ravel()
        A_s = raw.get_data(short_idx).ravel()

        # Eqn 27 Scholkmann et al 2014
        alfa = np.dot(A_s, A_l) / np.dot(A_s, A_s)

        # Eqn 26 Scholkmann et al 2014
        raw._data[pick] = A_l - alfa * A_s

    return raw


def _find_nearest_short(raw, pick, short_picks):
    """Return index of closest short channel.

    Parameters
    ----------
    raw : instance of Raw
        Haemoglobin data.
    pick : integer
        Index of channel of interest.
    short_picks : list of integers
        Indices of short channels.

    Returns
    -------
    pick : integer
        Index of short channel that is closest to pick.
        Note this index is relative to raw, and is from the subset provided
        in short_picks.

    """
    dist = [
        linalg.norm(raw.info["chs"][pick]["loc"][:3] - raw.info["chs"][p_sh]["loc"][:3])
        for p_sh in short_picks
    ]

    return short_picks[np.argmin(dist)]



# Authors: Robert Luke <mail@robertluke.net>
#
# License: BSD (3-clause)

import os

import matplotlib
import mne
import nilearn
import numpy as np
import pandas
import pytest
from matplotlib.pyplot import Axes
from mne.datasets import testing
from mne.fixes import _compare_version

from mne_nirs.experimental_design import make_first_level_design_matrix
from mne_nirs.statistics import RegressionResults, read_glm, run_glm

data_path = testing.data_path(download=False)
subjects_dir = data_path / "/subjects"


def _get_minimal_haemo_data(tmin=0, tmax=60):
    raw = mne.io.read_raw_nirx(
        os.path.join(mne.datasets.fnirs_motor.data_path(), "Participant-1"),
        preload=False,
    )
    raw.crop(tmax=tmax, tmin=tmin)
    raw = mne.preprocessing.nirs.optical_density(raw)
    raw = mne.preprocessing.nirs.beer_lambert_law(raw, ppf=0.1)
    raw.resample(0.3)
    raw.annotations.description[:] = [
        "e" + d.replace(".", "p") for d in raw.annotations.description
    ]
    return raw


def _get_glm_result(tmax=60, tmin=0, noise_model="ar1"):
    raw = _get_minimal_haemo_data(tmin=tmin, tmax=tmax)
    design_matrix = make_first_level_design_matrix(
        raw, stim_dur=5.0, drift_order=1, drift_model="polynomial"
    )
    return run_glm(raw, design_matrix, noise_model=noise_model)


def _get_glm_contrast_result(tmin=60, tmax=400):
    raw = _get_minimal_haemo_data(tmin=tmin, tmax=tmax)
    design_matrix = make_first_level_design_matrix(
        raw, stim_dur=5.0, drift_order=1, drift_model="polynomial"
    )

    glm_est = run_glm(raw, design_matrix)

    contrast_matrix = np.eye(design_matrix.shape[1])
    basic_conts = dict(
        [(column, contrast_matrix[i]) for i, column in enumerate(design_matrix.columns)]
    )
    assert "e1p" in basic_conts, sorted(basic_conts)
    contrast_LvR = basic_conts["e1p"] - basic_conts["e2p"]

    return glm_est.compute_contrast(contrast_LvR)


def test_create_results_glm():
    # Create a relevant info structure
    raw = _get_minimal_haemo_data()

    # Create a minimal structure
    res = _get_glm_result()

    # Get arguments for type so we can test errors below
    info = raw.info
    minimal_structure = res._data

    # Test construction
    with pytest.raises(TypeError, match="must be a dictionary"):
        _ = RegressionResults(info, np.zeros((5, 2)), 1)
    with pytest.raises(TypeError, match="must be a dictionary"):
        _ = RegressionResults(info, 3.2, 1)
    with pytest.raises(TypeError, match="must be a dictionary"):
        _ = RegressionResults(info, [], 1)
    with pytest.raises(TypeError, match="must be a dictionary"):
        _ = RegressionResults(info, "hbo", 1)
    with pytest.raises(TypeError, match="keys must match"):
        _ = RegressionResults(info, _take(4, minimal_structure), 1)

    onewrongname = _take(55, minimal_structure)
    onewrongname["test"] = onewrongname["S1_D1 hbo"]
    with pytest.raises(TypeError, match="must match ch_names"):
        _ = RegressionResults(info, onewrongname, 1)

    # Test properties
    assert len(res) == len(raw.ch_names)


def test_results_glm_properties():
    n_channels = 56

    res = _get_glm_result()

    # Test ContainsMixin
    assert "hbo" in res
    assert "hbr" in res
    assert "meg" not in res

    # Test copy
    assert len(res) == len(res.copy())

    # Test picks
    assert len(res.copy().pick(picks=range(4))) == 4
    assert len(res.copy().pick(picks="S1_D1 hbr")) == 1
    assert len(res.copy().pick(picks=["S1_D1 hbr"])) == 1
    assert len(res.copy().pick(picks=["S1_D1 hbr", "S1_D1 hbo"])) == 2

    if _compare_version(mne.__version__, "<", "1.4.0.dev140"):
        ctx = pytest.warns(RuntimeWarning, match="could not be picked")
    else:
        ctx = pytest.raises(ValueError, match="could not be picked")
    with ctx:
        assert len(res.copy().pick(picks=["S1_D1 hbr", "S1_D1 XXX"])) == 1

    assert len(res.copy().pick(picks="fnirs")) == n_channels
    assert len(res.copy().pick(picks="hbo")) == n_channels / 2
    assert len(res.copy().pick(picks="hbr")) == n_channels / 2

    # Test results
    assert len(res.theta()) == n_channels
    assert len(res.copy().pick(picks=range(4)).theta()) == 4
    assert len(res.copy().pick(picks=3).theta()) == 1
    assert res.copy().pick(picks=3).theta()[0].shape == (3, 1)

    # Test models
    assert len(res.model()) == n_channels
    assert len(res.copy().pick(picks=range(8)).model()) == 8
    assert type(res.model()[0]) is nilearn.glm.regression.ARModel

    assert isinstance(res.to_dataframe(), pandas.DataFrame)


def test_glm_scatter():
    assert isinstance(_get_glm_result().scatter(), Axes)
    assert isinstance(_get_glm_contrast_result().scatter(), Axes)


def test_results_glm_export_dataframe():
    n_channels = 56
    res = _get_glm_result(tmax=400)

    df = res.to_dataframe()

    assert df.shape == (6 * n_channels, 12)


def test_results_glm_export_dataframe_region_of_interest():
    res = _get_glm_result(tmax=400)

    # Create ROI with hbo only
    rois = dict()
    rois["A"] = [0, 2, 4]

    # Single ROI, single condition
    with pytest.raises(KeyError, match=r"not found in self\.design\.col"):
        res.to_dataframe_region_of_interest(rois, "1.0")
    df = res.to_dataframe_region_of_interest(rois, "e1p0")
    assert df.shape == (1, 9)
    assert df.Condition[0] == "e1p0"
    assert df.ROI[0] == "A"
    assert df.Chroma[0] == "hbo"

    # Single ROI, multiple conditions
    df = res.to_dataframe_region_of_interest(rois, ["e1p0", "e3p0", "drift_1"])
    assert df.shape == (3, 9)
    assert (df.Condition == ["e1p0", "e3p0", "drift_1"]).all()
    assert (df.ROI == ["A", "A", "A"]).all()

    # Single ROI, all conditions (default)
    df = res.to_dataframe_region_of_interest(rois, "e1p0")

    # HbR only ROI
    rois["B"] = [1, 3, 5]

    # Multiple ROI, single condition
    df = res.to_dataframe_region_of_interest(rois, "e1p0")
    assert df.shape == (2, 9)
    assert df.Condition[0] == "e1p0"
    assert df.Condition[1] == "e1p0"
    assert df.ROI[0] == "A"
    assert df.ROI[1] == "B"
    assert df.Chroma[0] == "hbo"
    assert df.Chroma[1] == "hbr"

    # Multiple ROI, multiple conditions
    df = res.to_dataframe_region_of_interest(rois, ["e1p0", "e3p0", "drift_1"])
    assert df.shape == (6, 9)

    # Multiple ROI, all conditions (default)
    df = res.to_dataframe_region_of_interest(rois, "e1p0")
    assert df.shape == (2, 9)

    # HbO and HbR ROI
    rois["C"] = [6, 7, 8]

    # Multiple ROI, single condition
    df = res.to_dataframe_region_of_interest(rois, "e1p0")
    assert df.shape == (4, 9)

    # Multiple ROI, multiple conditions
    df = res.to_dataframe_region_of_interest(rois, ["e1p0", "e3p0", "drift_1"])
    assert df.shape == (12, 9)

    # Multiple ROI, all conditions (default)
    df = res.to_dataframe_region_of_interest(rois, "e1p0")

    # With demographic information
    df = res.to_dataframe_region_of_interest(
        rois, ["e1p0", "e3p0", "drift_1"], demographic_info=True
    )
    assert df.shape == (12, 10)
    assert "Sex" in df.columns


def test_results_glm_export_dataframe_region_of_interest_weighted():
    res = _get_glm_result(tmax=400)
    res_df = res.to_dataframe().query("Condition == 'e1p0'")
    assert len(res_df)
    res_df["theta_uM"] = res_df["theta"] * 1e6

    # Create ROIs
    rois = dict()
    rois["A"] = [0, 2, 4, 6, 8, 10]
    rois["B"] = [1, 3, 5, 7, 9, 11]
    rois["C"] = [6, 7, 8, 9]

    df_uw = res.to_dataframe_region_of_interest(rois, "e1p0", weighted=False)
    assert df_uw.shape == (4, 9)
    assert df_uw.Weighted[0] == "Equal"
    thetas = np.array(res_df.theta)
    # unweighted option should be the same as a simple mean
    assert np.isclose(df_uw.query("ROI == 'A'").theta, thetas[rois["A"]].mean())

    df_w = res.to_dataframe_region_of_interest(rois, "e1p0", weighted=True)
    assert df_w.shape == (4, 9)
    assert df_w.Weighted[0] == "Inverse standard error"
    # weighted option should result in larger response
    # assert df_uw.query("ROI == 'A'").theta[0] < \
    #        df_w.query("ROI == 'A'").theta[0]

    # Create weights
    weights = dict()
    weights["A"] = [100000, 0.2, 0.4, 0.4, 0.4, 0.4]
    weights["B"] = [0.6, 0.1, 0.3, 10000.4, 10000.4, 0.4]
    weights["C"] = [16, 7, 8, 9]

    df = res.to_dataframe_region_of_interest(rois, "e1p0", weighted=weights)
    assert df.shape == (4, 9)
    assert df.Weighted[0] == "Custom"
    assert np.isclose(thetas[0], df.theta[0], atol=0.1e-6)
    assert np.isclose((thetas[7] + thetas[9]) / 2, df.theta[1], atol=0.1e-6)
    assert df.theta[2] > 0
    assert df.theta[3] < 0

    with pytest.raises(ValueError, match="must be positive"):
        weights["C"] = [16, 7, -8, 9]
        _ = res.to_dataframe_region_of_interest(rois, "e1p0", weighted=weights)

    with pytest.raises(ValueError, match="length of the keys"):
        weights["C"] = [16, 7]
        _ = res.to_dataframe_region_of_interest(rois, "e1p0", weighted=weights)

    with pytest.raises(KeyError, match="Keys of group_by and weighted"):
        bad_weights = dict()
        bad_weights["Z"] = [0, 2, 4]
        _ = res.to_dataframe_region_of_interest(rois, "e1p0", weighted=bad_weights)


def test_create_results_glm_contrast():
    # Create a minimal structure
    res = _get_glm_contrast_result()
    assert isinstance(res._data, nilearn.glm.contrasts.Contrast)
    assert isinstance(res.info, mne.Info)

    # Test copy
    assert len(res) == len(res.copy())

    assert isinstance(res.plot_topo(), matplotlib.figure.Figure)

    n_channels = 56
    assert isinstance(res.to_dataframe(), pandas.DataFrame)
    df = res.to_dataframe()
    assert df.shape == (n_channels, 10)

    src = np.asarray(df["Source"])
    assert src.dtype.kind == "i"


def test_results_glm_io():
    res = _get_glm_result(tmax=400)
    res.save("test-regression-glm.h5", overwrite=True)
    loaded_res = read_glm("test-regression-glm.h5")
    assert loaded_res.to_dataframe().equals(res.to_dataframe())
    assert res == loaded_res

    res = _get_glm_result(tmax=400, noise_model="ols")
    res.save("test-regression-ols_glm.h5", overwrite=True)
    loaded_res = read_glm("test-regression-ols_glm.h5")
    assert loaded_res.to_dataframe().equals(res.to_dataframe())
    assert res == loaded_res

    res = _get_glm_contrast_result()
    res.save("test-contrast-glm.h5", overwrite=True)
    loaded_res = read_glm("test-contrast-glm.h5")
    assert loaded_res.to_dataframe().equals(res.to_dataframe())
    assert res == loaded_res

    with pytest.raises(IOError, match="must end with glm.h5"):
        res.save("test-contrast-glX.h5", overwrite=True)


def _take(n, mydict):
    """Return first n items of the iterable as a list"""
    return {k: mydict[k] for k in list(mydict)[:n]}
